"""Applies electron recombination corrections."""

from typing import Optional, Union

import numpy as np

from spine.utils.globals import LAR_DENSITY, LAR_WION
from spine.utils.tracking import get_track_segment_dedxs

__all__ = ["RecombinationCalibrator"]


class RecombinationCalibrator:
    """Applies a recombination correction factor to account for some of the
    ionization electrons recombining with the Argon ions, which is an effect
    that depends on the local rate of energy deposition and the angle of
    the deposition trail (track) w.r.t. to the drift field.

    Notes
    -----
    Must call the gain calibrator upstream, which converts the number of ADCs
    to a number of observed ionization electrons.
    """

    name = "recombination"

    def __init__(
        self,
        efield: float,
        drift_dir: np.ndarray,
        model: str = "mbox",
        birks_a: float = 0.800,
        birks_k: float = 0.0486,
        mbox_alpha: float = 0.906,
        mbox_beta: float = 0.203,
        mbox_ell_r: float = 1.25,
        mip_dedx: float = 2.105168,
        tracking_mode: str = "bin_pca",
        **kwargs,
    ):
        """Initialize the recombination model and its constants.

        Parameters
        ----------
        efield : float
            Electric field in kV/cm
        drift_dir : np.ndarray
            (3) three-vector indicating the direction of the drift field
        model : str, default 'mbox'
            Recombination model name (one of 'birks', 'mbox' or 'mbox_ell')
        birks_a : float, default 0.800 (ICARUS CNGS fit)
            Birks model A parameter
        birks_k : float, default 0.0486 (ICARUS CNGS fit)
            Birks model k parameter in (kV/cm)(gm/cm^2)/MeV
        mbox_alpha : float, default 0.906 (ICARUS fit)
            Modified box model alpha parameter
        mbox_beta : float, default 0.203 (ICARUS fit)
            Modified box model beta parameter in (kV/cm)(g/cm^2)/MeV
        mbox_ell_r : float, default 1.25 (ICARUS fit)
            Modified box model ellipsoid correction R parameter
        mip_dedx : float, default 2.105168
            Mean dE/dx value of a MIP in LAr. Used to apply a flat recombination
            correction if the local dE/dx is not evaluated through tracking.
        track_mode : float, default 'bin_pca'
            If tracking is done to produce local dQ/dx values along tracks,
            defines the track chunking method to be used.
        **kwargs : dict, optional
            Additional arguments to pass to the tracking algorithm
        """
        # Store the drift direction
        self.drift_dir = drift_dir

        # Initialize the model parameters
        self.use_angles = False
        if model == "birks":
            self.model = "birks"
            self.a = birks_a
            self.k = birks_k / efield / LAR_DENSITY  # cm/MeV

        elif model in ["mbox", "mbox_ell"]:
            self.model = "mbox"
            self.alpha = mbox_alpha
            self.beta = mbox_beta / efield / LAR_DENSITY  # cm/MeV
            self.r = 1.0
            if model == "mbox_ell":
                self.use_angles = True
                self.r = mbox_ell_r

        else:
            raise ValueError(
                f"Recombination model not recognized: {model}. "
                "Must be one of 'birks', 'mbox' or 'mbox_ell'"
            )

        # Evaluate the MIP recombination factor, store it
        self.mip_recomb = self.recombination_factor(mip_dedx)

        # Store the tracking parameters
        self.tracking_mode = tracking_mode
        self.tracking_kwargs = kwargs

    def birks(self, dedx: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Birks equation to calculate electron quenching (higher local energy
        deposition are prone to more electron-ion recombination).

        Parameters
        ----------
        dedx : Union[float, np.ndarray]
            Value or array of values of dE/dx in MeV/cm

        Returns
        -------
        Union[float, np.ndarray]
           Quenching factors in electrons/MeV
        """
        return self.a / (1.0 + self.k * dedx)

    def inv_birks(self, dqdx: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Inverse Birks equation to undo electron quenching (higher local
        energy deposition are prone to more electron-ion recombination).

        Parameters
        ----------
        dqdx : Union[float, np.ndarray]
            Value or array of values of dQ/dx in electrons/cm

        Returns
        -------
        Union[float, np.ndarray]
            Inverse quenching factors in MeV/electrons
        """
        return 1.0 / (self.a / LAR_WION - self.k * dqdx)

    def mbox(
        self,
        dedx: Union[float, np.ndarray],
        cosphi: Optional[Union[float, np.ndarray]] = None,
    ) -> Union[float, np.ndarray]:
        """Modified box model equation to calculate electron quenching (higher
        local energy deposition are prone to more electron-ion recombination).

        Parameters
        ----------
        dedx : Union[float, np.ndarray]
            Value or array of values of dE/dx in MeV/cm
        cosphi : Union[float, np.ndarray], optional
            Value or array of values of the cosine of the angle w.r.t. the
            drift direction (in [0,1]).

        Returns
        -------
        Union[float, np.ndarray]
            Quenching factors in electrons/MeV
        """
        beta = self.beta
        if cosphi is not None:
            beta /= np.sqrt(1 - (1 - 1.0 / self.r**2) * cosphi**2)

        return np.log(self.alpha + beta * dedx) / beta / dedx

    def inv_mbox(
        self,
        dqdx: Union[float, np.ndarray],
        cosphi: Optional[Union[float, np.ndarray]] = None,
    ) -> Union[float, np.ndarray]:
        """Inverse modified box model equation to undo electron quenching
        (higher local energy deposition are prone to more electron-ion
        recombination).

        Parameters
        ----------
        dqdx : Union[float, np.ndarray]
            Value or array of values of dQ/dx in electrons/cm
        cosphi : Union[float, np.ndarray], optional
            Value or array of values of the cosine of the angle w.r.t. the
            drift direction (in [0,1]).

        Returns
        -------
        Union[float, np.ndarray]
            Inverse quenching factors in MeV/electrons
        """
        beta = self.beta
        if cosphi is not None:
            beta /= np.sqrt(1 - (1 - 1.0 / self.r**2) * cosphi**2)

        return (np.exp(beta * LAR_WION * dqdx) - self.alpha) / beta / dqdx

    def recombination_factor(
        self,
        dedx: Union[float, np.ndarray],
        cosphi: Optional[Union[float, np.ndarray]] = None,
    ) -> Union[float, np.ndarray]:
        """Calls the predefined recombination models to evaluate the
        appropriate quenching factors.

        Parameters
        ----------
        dedx : Union[float, np.ndarray]
            Value or array of values of dEdx in MeV/cm
        cosphi : Union[float, np.ndarray], optional
            Value or array of values of the cosine of the angle w.r.t. the
            drift direction (in [0,1]).

        Returns
        -------
        Union[float, np.ndarray]
            Quenching factors in electrons/MeV
        """
        if self.model == "birks":
            return self.birks(dedx)
        else:
            return self.mbox(dedx, cosphi)

    def inv_recombination_factor(
        self,
        dqdx: Union[float, np.ndarray],
        cosphi: Optional[Union[float, np.ndarray]] = None,
    ) -> Union[float, np.ndarray]:
        """Calls the predefined inverse recombination models to evaluate the
        appropriate correction factors.

        Parameters
        ----------
        dqdx : Union[float, np.ndarray]
            Value or array of values of dQ/dx in electrons/cm
        cosphi : Union[float, np.ndarray], optional
            Value or array of values of the cosine of the angle w.r.t. the
            drift direction (in [0,1]).

        Returns
        -------
        Union[float, np.ndarray]
            Inverse quenching factors in MeV/electrons
        """
        if self.model == "birks":
            return self.inv_birks(dqdx)
        else:
            return self.inv_mbox(dqdx, cosphi)

    def process(
        self,
        values: np.ndarray,
        points: Optional[np.ndarray] = None,
        track: bool = False,
    ) -> np.ndarray:
        """Corrects for electron recombination.

        Parameters
        ----------
        values : np.ndarray
            (N) array of depositions in number of electrons
        points : np.ndarray, optional
            (N, 3) array of point coordinates associated with one particle.
            Only needed if `track` is set to `True`.
        track : bool, defaut `False`
            Whether the object is a track or not. If it is, the track gets
            segmented to evaluate local dE/dx and track angle.

        Returns
        -------
        np.ndarray
            (N) array of depositions in MeV
        """
        # If no tracking is applied, use the MIP recombination factor
        if not track:
            return values * LAR_WION / self.mip_recomb

        # If the object is a track, segment the track use each segment to
        # compute a local dQ/dx (+ angle w.r.t. to the drift direction, if
        # requested) and assign a correction for all points in the segment.
        assert points is not None, "Cannot track the object without point coordinates"
        seg_dqdxs, _, _, seg_clusts, seg_dirs, _ = get_track_segment_dedxs(
            points, values, method=self.tracking_mode, **self.tracking_kwargs
        )

        corr_values = np.empty(len(values), dtype=values.dtype)
        for i, c in enumerate(seg_clusts):
            if seg_dqdxs[i] > 0.0:
                # If the dQ/dx was reliably estimated, use it
                if not self.use_angles:
                    corr = self.inv_recombination_factor(seg_dqdxs[i])
                else:
                    seg_cosphi = np.abs(np.dot(seg_dirs[i], self.drift_dir))
                    corr = self.inv_recombination_factor(seg_dqdxs[i], seg_cosphi)

            else:
                # If not, assume a MIP correction
                corr = LAR_WION / self.mip_recomb

            corr_values[c] = corr * values[c]

        return corr_values



"""Defines default plotly layouts."""

from copy import deepcopy

import matplotlib as mpl
import numpy as np
import seaborn as sns
from plotly import colors
from plotly import graph_objs as go
from plotly.subplots import make_subplots

from spine.geo import GeoManager

# Colorscale definitions
PLOTLY_COLORS = colors.qualitative.Plotly
PLOTLY_COLORS_TUPLE = colors.convert_colors_to_same_type(
    deepcopy(PLOTLY_COLORS), "tuple"
)[0]
PLOTLY_COLORS_WGRAY = ["#808080"] + PLOTLY_COLORS
HIGH_CONTRAST_COLORS = np.concatenate(
    [colors.qualitative.Dark24, colors.qualitative.Light24]
)

__all__ = ["layout3d", "dual_figure3d"]


def layout3d(
    ranges=None,
    meta=None,
    geo=None,
    use_geo=False,
    show_optical=False,
    show_crt=False,
    detector_padding=0.1,
    titles=None,
    detector_coords=False,
    backgroundcolor="white",
    gridcolor="lightgray",
    width=800,
    height=800,
    showlegend=True,
    camera=None,
    aspectmode="manual",
    aspectratio=None,
    dark=False,
    margin=None,
    hoverlabel=None,
    **kwargs,
):
    """Produces plotly.graph_objs.Layout object for a certain format.

    Parameters
    ----------
    ranges : np.ndarray, optional
        (3, 2) or (N, 3) Array used to specify the plot region in (x,y,z)
        directions. If not specified (None), the range will be set to include
        all points. Alternatively can be an array of shape (3,2) specifying
        (x,y,z) axis (min,max) range for a display, or simply a list of points
        with shape (N,3+) where [:,0],[:,1],[:,2] correspond to (x,y,z) values
        and the plotting region is decided by measuring the min,max range in
        each coordinates. This last option is useful if one wants to define
        the region based on a set of points that is not same as what's plotted.
    meta : Meta, optional
        Metadata information used to infer the full image range
    geo : Geometry, optional
        Geometry object used to define the detector boundaries
    use_geo : bool, default False
        If True, use the detector geometry to define the drawing range
    show_optical : bool, default False
        If True, the drawing range is extended to include the optical detectors
    show_crt : bool, default False
        If True, the drawing range is extended to include the CRT modules
    detector_padding : float, default 0.1
        Fraction of padding to add to the drawing range around the detector
    titles : List[str], optional
        (3) Array of strings for (x,y,z) axis title respectively
    detector_coords : bool, default False
        Whether or not the coordinates being drawn are in detector_coordinates
        or pixel IDs
    backgroundcolor : Union[str, int], default 'white'
        Color of the layout background
    gridcolor : Union[str, int], default 'lightgray'
        Color of the grid
    width : int, default 900
        Width of the layout in pixels
    height : int, default 900
        Height of the layout in pixels
    showlegend : bool, default True
        Whether or not to show the image legend
    aspectmode : str, default manual
        Plotly aspect mode. If manual, will define it based on the ranges
    aspectratio : dict, optional
        Plotly dictionary which specifies the aspect ratio for x, y an d z
    dark : bool, default False
        Dark layout
    margin : dict, optional
        Specifies the margin in each subplot
    hoverlabel : dict, optional
        Specifies the style hovertext labels
    **kwargs : dict, optional
        List of additional arguments to pass to plotly.graph_objs.Layout

    Results
    -------
    plotly.graph_objs.Layout
        Object that can be given to plotly.graph_objs.Figure for visualization
        (together with traces)
    """
    # Figure out the drawing range
    if ranges is None:
        ranges = [None, None, None]
    else:
        # If the range is provided, just use it
        if ranges.shape != (3, 2):
            assert (
                len(ranges.shape) == 2 and ranges.shape[1] == 3
            ), "If ranges is not of shape (3, 2), it must be of shape (N, 3)."
            ranges = np.vstack([np.min(ranges, axis=0), np.max(ranges, axis=0)]).T

        # Check that the range is sensible
        assert np.all(ranges[:, 1] >= ranges[:, 0])

    if use_geo or geo is not None:
        # If detector geometry is provided, make the full detector the range
        assert (
            ranges is None or None in ranges
        ), "Should not pass geo or ask to `use_geo=True` along with `ranges`."
        if geo is None:
            geo = GeoManager.get_instance()
        ranges = geo.get_boundaries(with_optical=show_optical, with_crt=show_crt)
        lengths = ranges[:, 1] - ranges[:, 0]

        # Add some padding
        ranges[:, 0] -= lengths * detector_padding
        ranges[:, 1] += lengths * detector_padding

        # If pixel coordinates are requested, use meta to make the conversion
        if detector_coords is False:
            assert meta is not None, (
                "Must provide metadata information to convert the detector "
                "coordinates to pixel coordinates."
            )
            ranges = meta.to_px(ranges.T).T

    elif meta is not None:
        # If meta information is provided, make the full image the range
        assert (
            ranges is None or None in ranges
        ), "Should not specify both `ranges` and `meta` parameters."
        if detector_coords:
            ranges = np.vstack([meta.lower, meta.upper]).T
        else:
            ranges = np.vstack(
                [[0, 0, 0], np.round((meta.upper - meta.lower) / meta.size)]
            ).T

    # Define detector-style camera, unless explicitely provided
    if camera is None:
        camera = {
            "eye": {"x": -2.0, "y": 1.0, "z": -0.01},
            "up": {"x": 0.0, "y": 1.0, "z": 0.0},
            "center": {"x": 0.0, "y": -0.1, "z": -0.01},
        }

    # Infer the image width/height and aspect ratios, unless they are specified
    if aspectmode == "manual":
        if aspectratio is None:
            axes = ["x", "y", "z"]
            ratios = np.ones(len(ranges))
            if ranges is not None and ranges[0] is not None:
                ranges_arr = np.array(ranges)
                assert ranges_arr.shape == (
                    3,
                    2,
                ), "If ranges is provided to infer aspectratio, it must be of shape (3, 2)."
                max_range = np.max(ranges_arr[:, 1] - ranges_arr[:, 0])
                ratios = (ranges_arr[:, 1] - ranges_arr[:, 0]) / max_range
            aspectratio = {axes[i]: v for i, v in enumerate(ratios)}

    # Check on the axis titles, define default
    assert titles is None or len(titles) == 3, "Must specify one title per axis"
    if titles is None:
        unit = "cm" if detector_coords else "pixel"
        titles = [f"x [{unit}]", f"y [{unit}]", f"z [{unit}]"]

    # Initialize some default legend behavior
    if "legend" not in kwargs:
        kwargs["legend"] = {
            "title": "Legend",
            "tracegroupgap": 1,
            "itemsizing": "constant",
        }

    # If a dark layout is requested, set the theme and the background color
    # accordingly
    if dark:
        kwargs["template"] = "plotly_dark"
        kwargs["paper_bgcolor"] = "black"
        backgroundcolor = "black"

    # If the margin is not provided, use 0 by default
    if margin is None:
        margin = {"b": 0, "t": 0, "l": 0, "r": 0}

    # Set hoverlabel font, if not provided
    if hoverlabel is None:
        hoverlabel = {"font_family": "Droid sans, monospace"}

    # Initialize the general scene layout
    axis_base = {
        "nticks": 10,
        "showticklabels": True,
        "tickfont": {"size": 14},
        "backgroundcolor": backgroundcolor,
        "gridcolor": gridcolor,
        "showbackground": True,
    }

    scene = {"aspectmode": aspectmode, "aspectratio": aspectratio, "camera": camera}
    for i, axis in enumerate(("x", "y", "z")):
        scene[f"{axis}axis"] = {
            "title": {"text": titles[i], "font": {"size": 20}},
            "range": ranges[i],
            **axis_base,
        }

    # Initialize layout
    layout = go.Layout(
        showlegend=showlegend,
        width=width,
        height=height,
        margin=margin,
        scene1=scene,
        scene2=deepcopy(scene),
        scene3=deepcopy(scene),
        hoverlabel=hoverlabel,
        **kwargs,
    )

    return layout


def dual_figure3d(
    traces_left,
    traces_right,
    layout=None,
    titles=None,
    width=1500,
    height=750,
    synchronize=False,
    margin=None,
    **kwargs,
):
    """Function which returns a plotly.graph_objs.Figure with two set of traces
    side-by-side in separate subplots.

    Parameters
    ----------
    traces_left : List[object]
        List of plotly traces to draw in the left subplot
    traces_right : List[object]
        List of plotly traces to draw in the right subplot
    layout : plotly.Layout, optional
        Plotly layout
    titles : List[str], optional
        Titles of the two subplots
    width : int, default 1000
        Width of the layout in pixels
    height : int, default 500
        Height of the layout in pixels
    synchronize : bool, default False
        If True, matches the camera position/angle of one plot to the other
    margin : dict, optional
        Specifies the margin in each subplot
    **kwargs : dict, optional
        List of additional arguments to pass to plotly.graph_objs.Layout

    Returns
    -------
    plotly.graph_objs.Figure
        Plotly figure with the two subplots
    """
    # If no title is provided, ommit them both
    if titles is None:
        titles = [None, None]

    # Make subplot and add traces
    fig = make_subplots(
        rows=1,
        cols=2,
        subplot_titles=(titles[0], titles[1]),
        specs=[[{"type": "scatter3d"}, {"type": "scatter3d"}]],
        horizontal_spacing=0.05,
        vertical_spacing=0.04,
    )

    num_left, num_right = len(traces_left), len(traces_right)
    fig.add_traces(traces_left, rows=[1] * num_left, cols=[1] * num_left)
    fig.add_traces(traces_right, rows=[1] * num_right, cols=[2] * num_right)

    # If the margin is not provided, use 20 by default
    if margin is None:
        margin = {"b": 20, "t": 20, "l": 20, "r": 20}

    # Inialize and set layout
    if layout is None:
        layout = layout3d(width=width, height=height, margin=margin, **kwargs)
    else:
        layout.update({"width": width, "height": height, "margin": margin})
    fig.layout.update(layout)

    # If requested, synchronize the two cameras
    if synchronize:
        fig = go.FigureWidget(fig)

        # Track if we're currently syncing to prevent infinite loops
        _syncing = [False]

        def cam_change_left(_, camera):
            if not _syncing[0]:
                _syncing[0] = True
                fig.layout.scene2.camera = camera
                _syncing[0] = False

        def cam_change_right(_, camera):
            if not _syncing[0]:
                _syncing[0] = True
                fig.layout.scene1.camera = camera
                _syncing[0] = False

        fig.layout.scene1.on_change(cam_change_left, "camera")
        fig.layout.scene2.on_change(cam_change_right, "camera")

    return fig


def apply_latex_style():
    """Sets the necessary :mod:`matplotlib` and :mod:`seaborn` parameters
    to draw a plot using latex style.
    """
    sns.set(
        rc={
            "figure.figsize": set_latex_size(250),
            "text.usetex": True,
            "font.family": "serif",
            "axes.labelsize": 8,
            "font.size": 8,
            "legend.fontsize": 8,
            "legend.labelspacing": 0.25,
            "legend.columnspacing": 0.25,
            "xtick.labelsize": 8,
            "ytick.labelsize": 8,
        },
        context="paper",
    )
    sns.set_style("white")
    sns.set_style(rc={"axes.grid": True, "font.family": "serif"})
    mpl.rcParams["text.latex.preamble"] = [r"\usepackage{amsmath,bm}"]


def set_latex_size(width, fraction=1):
    """Returns optimal figure dimension for a latex plot, depending on
    the requested width.

    Parameters
    ----------
    width : int
        Width of the page in points (pixels)
    fraction : float, default 1
        Fraction of the page width used by the figure

    Returns
    -------
    width : float
        Width of the figure in inches
    height : float
        Height of the figure in inches
    """
    # Width of figure (in pts)
    fig_width_pt = width * fraction

    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Golden ratio to set aesthetic figure height
    # https://disq.us/p/2940ij3
    golden_ratio = (5**0.5 - 1) / 2

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt

    # Figure height in inches
    fig_height_in = fig_width_in * golden_ratio

    return fig_width_in, fig_height_in


def color_rgba(color, alpha):
    """Convert an RGB color array into an RGBA string.

    Parameters
    ----------
    color : List[int]
        (3) List of RGB values
    alpha : float
        Alpha value in [0, 1]

    Returns
    -------
    str
        RGBA string
    """
    r, g, b = color
    return f"rgba({r}, {g}, {b}, {alpha})"



"""TPC detector geometry classes."""

from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional

import numpy as np

from .base import Box, Plane

__all__ = ["TPCDetector"]


@dataclass
class TPCChamber(Box):
    """Class which holds all properties of an individual time-projection
    chamber (TPC).

    Attributes
    ----------
    drift_dir : np.ndarray
        (3,) TPC drift direction vector (normalized)
    drift_axis : int
        Axis along which the electrons drift (0, 1 or 2)
    """

    drift_dir: np.ndarray
    drift_axis: int

    def __init__(
        self, position: np.ndarray, dimensions: np.ndarray, drift_dir: np.ndarray
    ):
        """Initialize the TPC object.

        Parameters
        ----------
        position : np.ndarray
            (3,) Position of the center of the TPC
        dimensions : np.ndarray
            (3,) Dimension of the TPC
        drift_dir : np.ndarray
            (3,) Drift direction vector
        """
        # Initialize the underlying box object
        lower = position - dimensions / 2
        upper = position + dimensions / 2
        super().__init__(lower, upper)

        # Make sure that the drift axis only points in one direction
        nonzero_axes = np.where(drift_dir)[0]
        assert (
            len(nonzero_axes) == 1
        ), "The drift direction must be aligned with a base axis."

        # Store drift information
        self.drift_dir = drift_dir
        self.drift_axis = nonzero_axes[0]

    @property
    def drift_sign(self) -> int:
        """Sign of drift w.r.t. to the drift axis orientation.

        Returns
        -------
        int
            Returns the sign of the drift vector w.r.t. to the drift axis
        """
        return int(self.drift_dir[self.drift_axis])

    @property
    def anode_side(self) -> int:
        """Returns whether the anode is on the lower or upper boundary of
        the TPC along the drift axis (0 for lower, 1 for upper).

        Returns
        -------
        int
            Anode side of the TPC
        """
        return (self.drift_sign + 1) // 2

    @property
    def cathode_side(self) -> int:
        """Returns whether the cathode is on the lower or upper boundary of
        the TPC along the drift axis (0 for lower, 1 for upper).

        Returns
        -------
        int
            Cathode side of the TPC
        """
        return 1 - self.anode_side

    @property
    def anode_pos(self) -> float:
        """Position of the anode along the drift direction.

        Returns
        -------
        float
            Anode position along the drift direction
        """
        return self.boundaries[self.drift_axis, self.anode_side]

    @property
    def cathode_pos(self) -> float:
        """Position of the cathode along the drift direction.

        Returns
        -------
        float
            Cathode position along the drift direction
        """
        return self.boundaries[self.drift_axis, self.cathode_side]


@dataclass
class TPCModule(Box):
    """Class which holds all properties of a TPC module.

    A module can hold either one chamber or two chambers with a shared cathode.

    Attributes
    ----------
    chambers : List[TPCChamber]
        List of individual TPCs that make up the module
    """

    chambers: List[TPCChamber]

    def __init__(
        self,
        positions: np.ndarray,
        dimensions: np.ndarray,
        drift_dirs: Optional[np.ndarray] = None,
    ):
        """Intialize the TPC module.

        Parameters
        ----------
        positions : np.ndarray
            (N_t) List of TPC center positions, one per TPC
        dimensions : np.ndarray
            (N_t) List of dimensions for each TPC
        drift_dirs : np.ndarray, optional
            (N_t, 3) List of drift directions. If this is not provided, it is
            inferred from the module configuration, provided that the module
            is composed of two TPCs with a shared cathode.
        """
        # Sanity checks
        assert len(positions) in (
            1,
            2,
            4,
        ), "A TPC module must contain exactly one, two, or four TPCs."
        assert (drift_dirs is not None) or (len(positions) == 2), (
            "For TPC modules with one or four TPC, the drift direction cannot be "
            "inferred and must be provided explicitely."
        )

        # Build TPCs
        self.chambers = []
        for t, position in enumerate(positions):
            # Fetch the drift axis. If not provided, join the two TPC centers
            if drift_dirs is not None:
                drift_dir = drift_dirs[t]
            else:
                drift_dir = position - positions[1 - t]
                drift_dir /= np.linalg.norm(drift_dir)

            # Instantiate TPC
            self.chambers.append(TPCChamber(positions[t], dimensions[t], drift_dir))

        # Initialize the underlying box object
        lower = np.min(np.vstack([c.lower for c in self.chambers]), axis=0)
        upper = np.max(np.vstack([c.upper for c in self.chambers]), axis=0)
        super().__init__(lower, upper)

    @property
    def num_chambers(self) -> int:
        """Number of individual TPCs that make up this module.

        Returns
        -------
        int
            Number of TPCs in the module
        """
        return len(self.chambers)

    @property
    def drift_axis(self) -> int:
        """Drift axis for the module (shared between chambers).

        Returns
        -------
        int
            Axis along which electrons drift in this module (0, 1 or 2)
        """
        return self.chambers[0].drift_axis

    @property
    def cathode_pos(self) -> float:
        """Location of the cathode plane along the drift axis.

        Returns
        -------
        float
            Location of the cathode plane along the drift axis
        """
        return np.mean([c.cathode_pos for c in self.chambers]).item()

    @property
    def cathode_thickness(self) -> float:
        """Thickness of the cathode.

        Returns
        -------
        float
            Thickness of the cathode
        """
        return abs(self.chambers[1].cathode_pos - self.chambers[0].cathode_pos)

    def __len__(self) -> int:
        """Returns the number of TPCs in the module.

        Returns
        -------
        int
            Number of TPCs in the module
        """
        return self.num_chambers

    def __getitem__(self, idx: int) -> TPCChamber:
        """Returns an underlying TPC of index idx.

        Parameters
        ----------
        idx : int
            Index of the TPC within the module

        Returns
        -------
        TPCChamber
            TPCChamber object
        """
        return self.chambers[idx]

    def __iter__(self) -> Iterator[TPCChamber]:
        """Resets an iterator counter, return self.

        Returns
        -------
        TPCModule
            The module itself
        """
        return iter(self.chambers)


@dataclass
class TPCDetector(Box):
    """Handles all geometry queries for a set of time-projection chambers.

    Attributes
    ----------
    modules : List[TPCModule]
        (N_m) List of TPC modules associated with this detector
    chambers : List[TPCChamber]
        (N_t) List of individual TPC associated with this detector
    det_ids : np.ndarray, optional
        (N_c) Map between logical and physical TPC index
    limits : List[Plane], optional
        (N_i) List of bounding planes which restrict the active volume
    """

    modules: List[TPCModule]
    chambers: List[TPCChamber]
    limits: Optional[List[Plane]] = None
    det_ids: Optional[np.ndarray] = None

    def __init__(
        self,
        dimensions: List[List[float]],
        positions: List[List[float]],
        module_ids: List[int],
        det_ids: Optional[List[int]] = None,
        drift_dirs: Optional[List[List[float]]] = None,
        limits: Optional[Dict[str, List[List[float]]]] = None,
    ):
        """Parse the detector boundary configuration.

        Parameters
        ----------
        dimensions : Union[List[float], List[List[float]]]
            (3/N_t) Dimensions of one TPC (if shared) or list of dimensions for each TPC
        positions : List[List[float]]
            (N_t) List of TPC center positions, one per TPC
        module_ids : List[int]
            (N_t) List of the module IDs each TPC belongs to
        det_ids : List[int], optional
            (N_c) Index of the physical detector which corresponds to each
            logical ID. This is needed if a TPC is divided into multiple logical
            IDs. If this is not specified, it assumed that there is a one-to-one
            correspondance between logical and physical.
        drift_dirs : List[List[float]], optional
            (N_t) List of drift direction vectors. If this is not provided, it
            is inferred from the module configuration, provided that modules
            are composed of two TPCs (with a shared cathode)
        limits: Dict[str, List[List[float]]], optional
            (N_i) Dictionary which defines a list of bounding planes which
            restrict the active region of the detector
        """
        # Check the sanity of the configuration
        assert np.all(
            [len(pos) == 3 for pos in positions]
        ), "Must provide the TPC position along 3 dimensions."
        assert len(module_ids) == len(
            positions
        ), "Must provide one module ID for each TPC."

        # If only one set of dimensions is provided, assume all TPCs share it
        if isinstance(dimensions[0], (float, int)):
            assert (
                len(dimensions) == 3
            ), "If only one set of dimensions is provided, it must be along 3 dimensions."
        else:
            assert len(dimensions) == len(positions), (
                "Must provide either one set of TPC dimensions for all TPCs, "
                "or one set of dimensions per TPC."
            )
            assert np.all(
                [len(dim) == 3 for dim in dimensions]
            ), "Each set of TPC dimensions must be along 3 dimensions."

        # Cast the dimensions, positions, ids to arrays
        module_ids_arr = np.asarray(module_ids, dtype=int)
        dimensions_arr = np.asarray(dimensions)
        if len(dimensions_arr.shape) == 1:
            dimensions_arr = np.tile(dimensions_arr, (len(positions), 1))
        positions_arr = np.asarray(positions)
        drift_dirs_arr = None
        if drift_dirs is not None:
            drift_dirs_arr = np.asarray(drift_dirs)

        # Construct TPC chambers, organized by module
        self.modules = []
        self.chambers = []
        module_drift_dirs = None
        for m in np.unique(module_ids):
            # Narrow down the set of TPCs to those in this module
            module_index = np.where(module_ids_arr == m)[0]
            module_positions = positions_arr[module_index]
            module_dimensions = dimensions_arr[module_index]
            if drift_dirs_arr is not None:
                module_drift_dirs = drift_dirs_arr[module_index]

            # Initialize the module, store
            module = TPCModule(module_positions, module_dimensions, module_drift_dirs)
            self.modules.append(module)
            self.chambers.extend(module.chambers)

        # Check that if detector IDs are provided, they are comprehensive
        if det_ids is not None:
            self.det_ids = np.asarray(det_ids, dtype=int)
            assert len(np.unique(det_ids)) == self.num_chambers_per_module, (
                "All physical TPCs must be associated with at least one " "logical TPC."
            )

        # Initialize the underlying all-encompasing box object
        lower = np.min(np.vstack([m.lower for m in self.modules]), axis=0)
        upper = np.max(np.vstack([m.upper for m in self.modules]), axis=0)
        super().__init__(lower, upper)

        # Initialize the active region limits
        if limits is not None:
            self.limits = self.initialize_limits(**limits)

    def initialize_limits(
        self, intercepts: List[List[float]], norms: List[List[float]]
    ) -> List[Plane]:
        """Initialize a list of bounding planes which restrict the active region.

        Parameters
        ----------
        intercepts : List[List[float]]
            List of plane intercept points
        norms : List[List[float]]
            List of vectors normal to the planes

        Returns
        -------
        List[Plane]
            List of bounding planes
        """
        limits = []
        for intercept, norm in zip(intercepts, norms):
            limits.append(Plane(np.asarray(intercept), np.asarray(norm)))

        return limits

    @property
    def num_chambers(self) -> int:
        """Number of individual TPC volumes.

        Returns
        -------
        int
            Number of TPC volumes, N_t
        """
        return len(self.chambers)

    @property
    def num_modules(self) -> int:
        """Number of detector modules.

        Returns
        -------
        int
            Number of detector modules, N_m
        """
        return len(self.modules)

    @property
    def num_chambers_per_module(self) -> int:
        """Number of TPC volumes per module.

        Returns
        -------
        int
            Number of TPC volumes per module, N_t/N_m
        """
        return len(self.modules[0])

    def __len__(self) -> int:
        """Returns the number of modules in the detector.

        Returns
        -------
        int
            Number of modules in the detector
        """
        return self.num_modules

    def __getitem__(self, idx: int) -> TPCModule:
        """Returns an underlying module of index idx.

        Parameters
        ----------
        idx : int
            Module index

        Returns
        -------
        TPCModule
            TPCModule object
        """
        return self.modules[idx]

    def __iter__(self) -> Iterator[TPCModule]:
        """Resets an iterator counter, return self.

        Returns
        -------
        TPCDetector
            The detector itself
        """
        return iter(self.modules)



"""Numba JIT compiled implementation of decomposition routines."""

import numba as nb
import numpy as np

__all__ = ["PCA", "principal_components"]

PCA_DTYPE = (("n_components", nb.int64),)


@nb.experimental.jitclass(PCA_DTYPE)
class PCA:
    """Class-version of the Numba-accelerate :func:`principal_components` function.

    Attributes
    ----------
    n_components : int
        Number of PCA components
    components : np.ndarray
        (N_c, D) List of principal axes
    explained_variance : np.ndarray
        (N_c) Variance along each of the principal axes
    """

    def __init__(self, n_components: nb.int64):
        """Initialize the PCA parameters.

        Parameters
        ----------
        n_components : int
            Number of PCA components, N_c
        """
        # Store parameters
        assert n_components > 0, "Must at least include one component."
        self.n_components = n_components

    def fit(self, x):
        """Computes the covariance and eigen-decompose the data.

        Parameters
        ----------
        x : np.ndarray
            (N, D) array of point coordinates in some D-dimensional space

        Returns
        -------
        components : np.ndarray
            (N_c, D) List of principal axes
        explained_variance : np.ndarray
            (N_c) Variance along each of the principal axes
        """
        # Check input
        assert x.shape[1] >= self.n_components, (
            f"The dimensionality of the data ({x.shape[1]}) is smaller "
            f"than the number of components ({self.n_components}."
        )

        # Compute the covariance matrix
        A = np.cov(x.T, ddof=len(x) - 1).astype(x.dtype)

        # Eigen-decompose the covariance matrix
        # TODO: get rid of casting, this is a complex LAPACK issue currently
        w, v = np.linalg.eigh(A.astype(np.float64))
        w, v = w.astype(x.dtype), v.astype(x.dtype)
        w, v = np.flip(w), np.ascontiguousarray(np.fliplr(v).T)

        # Store output
        return v[: self.n_components], w[: self.n_components] / (len(x) - 1)


@nb.njit(cache=True)
def principal_components(x: nb.float32[:, :]) -> nb.float32[:, :]:
    """Computes the principal components of a point cloud by computing the
    eigenvectors of the centered covariance matrix.

    Parameters
    ----------
    x : np.ndarray
        (N, d) Coordinates in d dimensions

    Returns
    -------
    np.ndarray
        (d, d) List of principal components (row-ordered)
    """
    # Get covariance matrix
    A = np.cov(x.T, ddof=len(x) - 1).astype(x.dtype)  # Casting needed...

    # Get eigenvectors
    # TODO: get rid of casting, this is a complex LAPACK issue currently
    _, v = np.linalg.eigh(A.astype(np.float64))
    v = v.astype(x.dtype)
    v = np.ascontiguousarray(np.fliplr(v).T)

    return v



"""File download utilities for SPINE configuration system.

This module provides utilities for downloading model weights and other files
referenced in YAML configurations.
"""

import hashlib
import os
import sys
import urllib.parse
import urllib.request
from pathlib import Path
from typing import Optional
from urllib.error import HTTPError, URLError


def get_cache_dir() -> Path:
    """Get the directory for caching downloaded files.

    By default, creates a '.cache/weights/' directory in SPINE_BASEDIR or
    SPINE_PROD_BASEDIR (if running from spine-prod). This ensures downloads
    are cached centrally regardless of execution directory.

    Can be overridden with the SPINE_CACHE_DIR environment variable.

    Returns
    -------
    Path
        Path to cache directory
    """
    cache_dir = os.environ.get("SPINE_CACHE_DIR")
    if cache_dir:
        return Path(cache_dir)

    # Try SPINE_PROD_BASEDIR first (if running from spine-prod)
    spine_prod_base = os.environ.get("SPINE_PROD_BASEDIR")
    if spine_prod_base:
        return Path(spine_prod_base) / ".cache" / "weights"

    # Fall back to SPINE_BASEDIR
    spine_base = os.environ.get("SPINE_BASEDIR")
    if spine_base:
        return Path(spine_base) / ".cache" / "weights"

    # Last resort: use current directory (with warning)
    print(
        "WARNING: SPINE_BASEDIR and SPINE_PROD_BASEDIR not set. "
        "Using current directory for cache. "
        "Please source configure.sh for proper caching."
    )
    return Path.cwd() / "weights"


def compute_file_hash(filepath: Path, algorithm: str = "sha256") -> str:
    """Compute hash of a file.

    Parameters
    ----------
    filepath : Path
        Path to file to hash
    algorithm : str, optional
        Hash algorithm to use (default: sha256)

    Returns
    -------
    str
        Hex digest of file hash
    """
    hasher = hashlib.new(algorithm)
    with open(filepath, "rb") as f:
        while True:
            chunk = f.read(8192)
            if not chunk:
                break
            hasher.update(chunk)
    return hasher.hexdigest()


def url_to_filename(url: str) -> str:
    """Convert URL to a safe filename using hash.

    Preserves the original extension if present.

    Parameters
    ----------
    url : str
        URL to convert

    Returns
    -------
    str
        Safe filename
    """
    # Extract extension from URL
    url_path = urllib.parse.urlparse(url).path
    ext = Path(url_path).suffix

    # Hash the URL to create a unique filename
    url_hash = hashlib.sha256(url.encode()).hexdigest()[:16]

    # Combine hash with extension
    return f"{url_hash}{ext}"


def download_file(
    url: str,
    output_path: Path,
    expected_hash: Optional[str] = None,
    hash_algorithm: str = "sha256",
) -> None:
    """Download a file from a URL with progress reporting.

    Parameters
    ----------
    url : str
        URL to download from
    output_path : Path
        Path where file should be saved
    expected_hash : str, optional
        Expected hash of downloaded file for validation
    hash_algorithm : str, optional
        Hash algorithm to use for validation (default: sha256)

    Raises
    ------
    HTTPError
        If download fails with HTTP error
    URLError
        If download fails with URL error
    ValueError
        If downloaded file hash doesn't match expected hash
    """
    print(f"Downloading: {url}")
    print(f"Destination: {output_path}")

    try:
        # Download with progress reporting
        def progress_hook(count, block_size, total_size):
            if total_size > 0:
                percent = min(100, count * block_size * 100 // total_size)
                sys.stdout.write(f"\rProgress: {percent}% ")
                sys.stdout.flush()

        urllib.request.urlretrieve(url, output_path, reporthook=progress_hook)
        print()  # New line after progress

    except HTTPError as e:
        raise HTTPError(
            url, e.code, f"HTTP Error {e.code}: {e.reason}", e.hdrs, e.fp
        ) from e
    except URLError as e:
        raise URLError(f"Failed to download {url}: {e.reason}") from e

    # Validate hash if provided
    if expected_hash:
        actual_hash = compute_file_hash(output_path, hash_algorithm)
        if actual_hash != expected_hash:
            output_path.unlink()  # Remove corrupted file
            raise ValueError(
                f"Downloaded file hash mismatch!\n"
                f"Expected: {expected_hash}\n"
                f"Got:      {actual_hash}\n"
                f"File has been removed. Please try again."
            )
        print(f"✓ Hash validated: {actual_hash[:16]}...")


def download_from_url(
    url: str,
    expected_hash: Optional[str] = None,
    cache_dir: Optional[Path] = None,
) -> str:
    """Download a file from URL and return the cached path.

    If the file already exists in cache and passes hash validation (if provided),
    returns the cached path without re-downloading.

    Parameters
    ----------
    url : str
        URL to download from
    expected_hash : str, optional
        Expected SHA256 hash of file for validation
    cache_dir : Path, optional
        Directory to cache downloaded files (default: from get_cache_dir())

    Returns
    -------
    str
        Absolute path to cached file

    Raises
    ------
    HTTPError
        If download fails with HTTP error
    URLError
        If download fails with URL error
    ValueError
        If hash validation fails

    Examples
    --------
    >>> path = download_from_url(
    ...     "https://example.com/model.ckpt",
    ...     expected_hash="abc123..."
    ... )
    >>> print(f"Model downloaded to: {path}")
    """
    if cache_dir is None:
        cache_dir = get_cache_dir()

    # Create cache directory if it doesn't exist
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Generate filename from URL
    filename = url_to_filename(url)
    output_path = cache_dir / filename

    # Check if file already exists
    if output_path.exists():
        # Validate existing file if hash provided
        if expected_hash:
            try:
                actual_hash = compute_file_hash(output_path)
                if actual_hash == expected_hash:
                    print(f"✓ Using cached file: {output_path}")
                    return str(output_path.absolute())
                else:
                    print("⚠ Cached file hash mismatch, re-downloading...")
                    output_path.unlink()

            except (OSError, IOError, ValueError) as e:
                print(f"⚠ Error validating cached file: {e}, re-downloading...")
                output_path.unlink()
        else:
            print(f"✓ Using cached file: {output_path}")
            return str(output_path.absolute())

    # Download the file
    download_file(url, output_path, expected_hash)

    return str(output_path.absolute())



"""Contains the data reader base class.

Data readers are used to extract specific entries from files and store their
data products into dictionaries to be used downstream.
"""

import glob
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from spine.utils.logger import logger


class ReaderBase(ABC):
    """Parent reader class which provides common functions between all readers.

    This class provides these basic functions:
    1. Method to parse the requested file list or file list file into a list of
       paths to existing files (throws if nothing is found)
    2. Method to produce a list of entries in the file(s) as selected by the
       provided parameters, checks that they exist (throws if they do not)
    3. Essential `__len__` and `__getitem__` methods. Must define the
       `get` function in the inheriting class for both of them to work.

    Attributes
    ----------
    name : str
        Name of the reader, as requested in the configuration
    num_entries : int
        Total number of entries in the files provided
    entry_index : List[int]
        List of global indexes to cycle through
    file_paths : List[str]
        List of files to read data from
    file_offsets : List[int]
        Offsets between the global index and each individual file start index
    file_index : List[int]
        Index of the file each entry in entry_index lives in
    run_info : np.ndarray
        (run, subrun, event) triplets associated with each entry in the file list
    run_map : Dict[Tuple[int], int]
        Maps each available (run, subrun, event) triplet onto an entry_index index
    """

    name: str
    file_paths: List[str]
    file_index: np.ndarray
    file_offsets: np.ndarray
    entry_index: np.ndarray
    num_entries: int
    run_info: Optional[List[Tuple[int, int, int]]] = None
    run_map: Optional[Dict[Tuple[int, int, int], int]] = None

    def __len__(self) -> int:
        """Returns the number of entries in the file(s).

        Returns
        -------
        int
            Number of entries in the file
        """
        return len(self.entry_index)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """Returns a specific entry in the file.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        dict
            One entry-worth of data from the loaded files
        """
        return self.get(idx)

    @abstractmethod
    def get(self, idx: int) -> Dict[str, Any]:
        """Placeholder to be defined by the daughter class."""
        raise NotImplementedError

    def get_run_event(self, run: int, subrun: int, event: int) -> Dict[str, Any]:
        """Returns an entry corresponding to a specific (run, subrun, event)
        triplet.

        Parameters
        ----------
        run : int
            Run number
        subrun : int
            Subrun number
        event : int
            Event number

        Returns
        -------
        data_blob : dict
            Ditionary of input data products corresponding to one event
        """
        return self.get(self.get_run_event_index(run, subrun, event))

    def process_file_paths(
        self,
        file_keys: Optional[Union[str, List[str]]] = None,
        file_list: Optional[str] = None,
        limit_num_files: Optional[int] = None,
        max_print_files: int = 10,
    ) -> None:
        """Process list of files.

        Parameters
        ----------
        file_keys : Union[str, List[str]], optional
            Path or list of paths to the files to be read
        file_list : str, optional
            Path to a text file containing a list of file paths to be read
        limit_num_files : int, optional
            Integer limiting number of files to be taken per data directory
        max_print_files : int, default 10
            Maximum number of loaded file names to be printed
        """
        # Some basic checks
        assert (file_keys is not None) ^ (
            file_list is not None
        ), "Must provide either `file_keys` or `file_list` to process files."
        assert (
            limit_num_files is None or limit_num_files > 0
        ), "If `limit_num_files` is provided, it must be larger than 0."

        # When using a file list in text format, read it and parse to a list
        if file_list is not None:
            # If the file list is a text file, extract the list of paths
            assert os.path.isfile(file_list), (
                "If the `file_list` arguments is provided, it must point to a valid "
                "path to a text file with a file list."
            )
            with open(file_list, "r", encoding="utf-8") as f:
                file_keys = f.read().splitlines()

        # Convert the file keys to a list of file paths with glob
        assert file_keys is not None  # For the linter's sake
        self.file_paths = []
        if isinstance(file_keys, str):
            file_keys = [file_keys]
        for file_key in file_keys:
            file_paths = glob.glob(file_key)
            assert file_paths, f"File key {file_key} yielded no compatible path."
            for path in file_paths:
                if (
                    limit_num_files is not None
                    and len(self.file_paths) >= limit_num_files
                ):
                    break
                self.file_paths.append(path)

        self.file_paths = sorted(self.file_paths)

        # Print out the list of loaded files
        num_files = len(self.file_paths)
        file_list = " - " + "\n - ".join(self.file_paths[:max_print_files])
        file_list += "\n ... \n" if num_files > max_print_files else "\n"
        logger.info("Will load %d file(s):\n%s", num_files, file_list)

    def process_run_info(self) -> None:
        """Process the run information.

        Check the run information for duplicates and initialize a dictionary
        which map (run, subrun, event) triplets onto entry index.
        """
        # Check for duplicates
        if self.run_info is not None:
            assert len(self.run_info) == self.num_entries
            num_unique = len(np.unique(self.run_info, axis=0))
            assert num_unique == len(self.run_info), (
                "Cannot create a run map if (run, subrun, event) triplets "
                "are not unique in the dataset. Abort."
            )

        # If run_info is set, flip it into a map from info to entry
        self.run_map = None
        if self.run_info is not None:
            self.run_map = {v: i for i, v in enumerate(self.run_info)}

    def process_entry_list(
        self,
        n_entry: Optional[int] = None,
        n_skip: Optional[int] = None,
        entry_list: Optional[Union[str, List[int]]] = None,
        skip_entry_list: Optional[Union[str, List[int]]] = None,
        run_event_list: Optional[Union[str, List[List[int]]]] = None,
        skip_run_event_list: Optional[Union[str, List[List[int]]]] = None,
        allow_missing: bool = False,
    ) -> None:
        """Create a list of entries that can be accessed by :meth:`__getitem__`.

        Parameters
        ----------
        n_entry : int, optional
            Maximum number of entries to load
        n_skip : int, optional
            Number of entries to skip at the beginning
        entry_list : Union[str, List[int]], optional
            List of integer entry IDs to add to the index. Can be passed
            as a file path or a list of integers
        skip_entry_list : Union[str, List[int]], optional
            List of integer entry IDs to skip from the index. Can be passed
            as a file path or a list of integers
        run_event_list: Union[str, List[List[int]]], optional
            List of (run, subrun, event) triplets to add to the index. Can be passed
            as a file path or a list of triplets
        skip_run_event_list: Union[str, List[List[int]]], optional
            List of (run, subrun, event) triplets to skip from the index. Can be passed
            as a file path or a list of triplets
        allow_missing : bool, default False
            If `True`, allows missing entries in the entry or event list

        Returns
        -------
        list
            List of integer entry IDs in the index
        """
        # Make sure the parameters are sensible
        if np.any(
            [
                n_entry is not None,
                n_skip is not None,
                entry_list is not None,
                skip_entry_list is not None,
                run_event_list is not None,
                skip_run_event_list is not None,
            ]
        ):
            assert (
                (n_entry is not None or n_skip is not None)
                ^ (entry_list is not None or skip_entry_list is not None)
                ^ (run_event_list is not None or skip_run_event_list is not None)
            ), (
                "Cannot specify `n_entry` or `n_skip` at the same time "
                "as `entry_list` or `skip_entry_list` or at the same time "
                "as `run_event_list` or `skip_run_event_list`."
            )

        if n_entry is not None or n_skip is not None:
            n_skip = n_skip if n_skip else 0
            n_entry = n_entry if n_entry else self.num_entries - n_skip
            assert n_skip + n_entry <= self.num_entries, (
                f"Incompatibility between `n_entry` ({n_entry}), `n_skip` ({n_skip}) "
                f"and the number of entries in the files ({self.num_entries})."
            )

        assert not entry_list or not skip_entry_list, (
            "Cannot specify both `entry_list` and "
            "`skip_entry_list` at the same time."
        )

        assert not run_event_list or not skip_run_event_list, (
            "Cannot specify both `run_event_list` and "
            "`skip_run_event_list` at the same time."
        )

        # Create a list of entries to be loaded
        entry_list_arr = None
        if n_entry or n_skip:
            entry_list_arr = np.arange(self.num_entries)
            if n_skip is not None and n_skip > 0:
                entry_list_arr = entry_list_arr[n_skip:]
            if n_entry is not None and n_entry > 0:
                entry_list_arr = entry_list_arr[:n_entry]

        elif entry_list:
            entry_list_arr = self.parse_entry_list(entry_list)
            assert np.all(
                entry_list_arr < self.num_entries
            ), "Values in entry_list outside of bounds."

        elif run_event_list:
            assert (
                self.run_map is not None
            ), "Must build a run map to get entries by (run, subrun, event)."
            self.process_run_info()
            run_event_list_tuple = self.parse_run_event_list(run_event_list)
            entry_list = []
            for r, s, e in run_event_list_tuple:
                if not allow_missing or (r, s, e) in self.run_map:
                    entry_list.append(self.get_run_event_index(r, s, e))

            entry_list_arr = np.unique(entry_list)

        elif skip_entry_list or skip_run_event_list:
            skip_entry_list_arr = None
            if skip_entry_list:
                skip_entry_list_arr = self.parse_entry_list(skip_entry_list)
                assert np.all(
                    skip_entry_list_arr < self.num_entries
                ), "Values in skip_entry_list outside of bounds."

            elif skip_run_event_list:
                assert (
                    self.run_map is not None
                ), "Must build a run map to skip entries by (run, subrun, event)."
                self.process_run_info()
                skip_run_event_list_tuple = self.parse_run_event_list(
                    skip_run_event_list
                )
                skip_entry_list = []
                for r, s, e in skip_run_event_list_tuple:
                    if not allow_missing or (r, s, e) in self.run_map:
                        skip_entry_list.append(self.get_run_event_index(r, s, e))

                skip_entry_list_arr = np.array(skip_entry_list, dtype=int)

            if skip_entry_list_arr is not None:
                entry_mask = np.ones(self.num_entries, dtype=bool)
                entry_mask[skip_entry_list_arr] = False
                entry_list_arr = np.where(entry_mask)[0]

        # Apply entry list to the indexes
        entry_index = np.arange(self.num_entries, dtype=np.int64)
        if entry_list_arr is not None:
            entry_index = entry_index[entry_list_arr]

            if self.run_info is not None:
                run_info = [self.run_info[i] for i in entry_list_arr]
                self.run_map = {v: i for i, v in enumerate(run_info)}

        assert len(entry_index), "Must at least have one entry to load."

        logger.info("Total number of entries selected: %d\n", len(entry_index))

        self.entry_index = entry_index

    def get_run_event_index(self, run: int, subrun: int, event: int) -> int:
        """Returns an entry index corresponding to a specific
        (run, subrun, event) triplet.

        Parameters
        ----------
        run : int
            Run number
        subrun : int
            Subrun number
        event : int
            Event number
        """
        # Get the appropriate entry index
        assert (
            self.run_map is not None
        ), "Must build a run map to get entries by (run, sunrun, event)."
        assert (
            run,
            subrun,
            event,
        ) in self.run_map, (
            f"Could not find (run={run}, subrun={subrun}, event={event})."
        )

        return self.run_map[(run, subrun, event)]

    def get_file_path(self, idx: int) -> str:
        """Returns the path to the file corresponding to a specific entry.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        str
            Path to the file
        """
        return self.file_paths[self.get_file_index(idx)]

    def get_file_index(self, idx: int) -> int:
        """Returns the index of the file corresponding to a specific entry.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        int
            Index of the file in the file list
        """
        return int(self.file_index[self.entry_index[idx]])

    def get_file_entry_index(self, idx: int) -> int:
        """Returns the index of an entry within the file it lives in,
        provided a global index over the list of files.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        int
            Index of the entry in the file
        """
        file_idx = self.get_file_index(idx)
        offset = self.file_offsets[file_idx]
        return int(self.entry_index[idx] - offset)

    @staticmethod
    def parse_entry_list(list_source: Union[str, List[int]]) -> np.ndarray:
        """Parses a list into an np.ndarray.

        The list can be passed as a simple python list or a path to a file
        which contains space or comma separated numbers (can be on multiple
        lines or not)

        Parameters
        ----------
        list_source : Union[str, List[int]]
            List as a python list or a text file path

        Returns
        -------
        np.ndarray
            List as a numpy array
        """
        if list_source is None:
            return np.empty(0, dtype=np.int64)

        if not np.isscalar(list_source):
            return np.asarray(list_source, dtype=int)

        if isinstance(list_source, str):
            assert os.path.isfile(list_source), "The list source file does not exist."
            with open(list_source, "r", encoding="utf-8") as f:
                lines = f.read().splitlines()
                line_list = [l.replace(",", " ").split() for l in lines]
                list_source = [int(w) for l in line_list for w in l]

            return np.array(list_source, dtype=int)

        raise ValueError("List format not recognized. Must be a list or a file path.")

    @staticmethod
    def parse_run_event_list(
        list_source: Union[str, List[List[int]]],
    ) -> List[Tuple[int, int, int]]:
        """Parses a list of (run, subrun, event) triplets into an np.ndarray.

        The list can be passed as a simple python list or a path to a file
        which contains one (run, subrun, event) pair per line.

        Parameters
        ----------
        list_source : Union[str, List[List[int]]]
            List as a python list or a text file path

        Returns
        -------
        List[Tuple[int, int, int]]
            List as a numpy array
        """
        if list_source is None:
            return []

        if isinstance(list_source, list):
            assert all(
                len(val) == 3 for val in list_source
            ), "Each entry in the run-event list must have three integers."
            return [(int(r), int(s), int(e)) for r, s, e in list_source]

        if isinstance(list_source, str):
            assert os.path.isfile(list_source), "The list source file does not exist."
            with open(list_source, "r", encoding="utf-8") as f:
                lines = f.read().splitlines()
                line_list = [l.replace(",", " ").split() for l in lines]
                assert all(
                    len(l) == 3 for l in line_list
                ), "Each line in the run-event list file must have three integers."
                return [(int(r), int(s), int(e)) for r, s, e in line_list]

        raise ValueError("List format not recognized.")



"""Module with helper functions to run inference on a model configuration."""

from copy import deepcopy

import yaml


def get_inference_cfg(
    cfg, file_keys=None, weight_path=None, batch_size=None, num_workers=None, cpu=False
):
    """Turns a training configuration into an inference configuration.

    This script does the following:
    - Turn `train` to `False`
    - Set sequential sampling
    - Load the specified validation file_keys, if requested
    - Load the specified set of weight_path, if requested
    - Reset the batch_size to a different value, if requested
    - Sets num_workers to a different value, if requested
    - Make the model run in CPU mode, if requested

    Parameters
    ----------
    cfg : Union[str, dict]
        Configuration file or Path to the configuration file
    file_keys : str, optional
        Path to the dataset to use for inference
    weight_path : str, optional
        Path to the weigths to use for inference
    batch_size : int, optional
        Number of data samples per batch
    num_workers : int, optional
        Number of workers that load data
    cpu : bool, default False
        Whether or not to execute the inference on CPU

    Returns
    -------
    dict
        Dictionary of parameters to initialize handlers
    """
    # Fetch the training configuration
    if isinstance(cfg, dict):
        cfg = deepcopy(cfg)
    else:
        cfg = open(cfg, "r", encoding="utf-8")
        cfg = yaml.safe_load(cfg)

    # Turn train to False
    if "train" in cfg["base"]:
        del cfg["base"]["train"]

    # Turn on unwrapper
    cfg["base"]["unwrap"] = True

    # Convert mode output to numpy
    cfg["model"]["to_numpy"] = True

    # Get rid of random sampler
    if "sampler" in cfg["io"]["loader"]:
        del cfg["io"]["loader"]["sampler"]

    # Change the batch_size, if requested
    if batch_size is not None:
        cfg["io"]["loader"]["batch_size"] = batch_size

    # Change dataset, if requested
    if file_keys is not None:
        cfg["io"]["loader"]["dataset"]["file_keys"] = file_keys

    # Set the number of workers, if requested
    if num_workers is not None:
        cfg["io"]["loader"]["num_workers"] = num_workers

    # Change weights, if requested
    if weight_path is not None:
        cfg["model"]["weight_path"] = weight_path

    # Put the network in CPU mode, if requested
    if cpu:
        cfg["base"]["world_size"] = 0

    return cfg



"""Typed exceptions for SPINE configuration loading.

This module defines specific exception types for different kinds of
configuration errors, making it easier to handle and debug issues.
"""

from typing import List


class ConfigError(Exception):
    """Base exception for all configuration errors."""


class ConfigIncludeError(ConfigError):
    """Raised when an included file cannot be found or loaded."""


class ConfigCycleError(ConfigError):
    """Raised when a circular include dependency is detected."""

    def __init__(self, cycle_path: List[str]):
        """Initialize with the cycle path.

        Parameters
        ----------
        cycle_path : List[str]
            List of file paths showing the include cycle
        """
        self.cycle_path = cycle_path
        cycle_str = " -> ".join(cycle_path)
        super().__init__(f"Circular include detected: {cycle_str}")


class ConfigPathError(ConfigError):
    """Raised when a configuration path cannot be resolved or does not exist."""


class ConfigTypeError(ConfigError):
    """Raised when a configuration operation is applied to the wrong type."""


class ConfigOperationError(ConfigError):
    """Raised when a configuration operation (e.g., collection op) fails."""


class ConfigValidationError(ConfigError):
    """Raised when configuration validation fails (e.g., compatibility checks)."""



"""Tools to draw voxelized data organized in clusts."""

import numpy as np

from spine.math.distance import closest_pair
from spine.utils.globals import COORD_COLS

from .cluster import scatter_clusters
from .point import scatter_points

__all__ = ["network_topology", "network_schematic"]


def network_topology(
    points,
    clusts,
    edge_index,
    clust_labels=None,
    edge_labels=None,
    mode="scatter",
    color=None,
    line=None,
    linewidth=2,
    name=None,
    **kwargs,
):
    """Network 3D topological representation in Euclidean space.

    Parameters
    ----------
    points : np.ndarray
        (N, 3) array of N points of (..., x, y, z,...) coordinate information
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    edge_index : np.ndarray
        (E, 2) List of connections between clusters
    clust_labels : np.ndarray, optional
        (C) List of cluster labels
    edge_labels : np.ndarray, optional
        (E) List of edge labels
    mode : str, default 'scatter'
        Drawing mode; one of 'circle', 'scatter', 'ellipsoid', 'cone' or 'hull'
    color : Union[str, np.ndarray], optional
        Color of clusters or (C) list of color of clusters
    line : dict, optional
        Line property dictionary
    linewidth : float, default 2
        Width of the edge lines
    name : str, optional
        Name of the network
    **kwargs : dict, optional
        List of additional arguments to pass to plotly

    Returns
    -------
    List[Union[plotly.graph_objs.Scatter3d, plotly.graph_objs.Mesh3d]]
        Node and edge traces in the same list
    """
    # Fetch the list of point coordinates
    if points.shape[1] != 3:
        points = points[:, COORD_COLS]

    # Check that color is not passed directly, ambiguous for a network
    assert (
        color is None
    ), "Use `clust_labels` instead of `color` to specify node colors."

    # Set the prefix to add to the trace names
    prefix = f"{name}" if name is not None else "Graph"
    node_name = name if edge_index is None else f"{prefix} nodes"
    edge_name = f"{prefix} edges"

    # Define the trace(s) associated with the graph nodes
    single_trace = mode in ["circle", "scatter"]
    traces = scatter_clusters(
        points,
        clusts,
        color=clust_labels,
        single_trace=single_trace,
        name=node_name,
        mode=mode,
        **kwargs,
    )

    # Define the trace associated with graph edges
    edge_vertices = np.empty((0, 3), dtype=points.dtype)
    if len(edge_index):
        edge_vertices = []
        if mode in ["circle", "ellipsoid"]:
            # For circles and ellipsoids, join centroid to centroid
            cent = [points[c].mean(axis=0) for c in clusts]
            for i, j in edge_index:
                edge_vertices.extend([cent[i], cent[j], [None, None, None]])

        elif mode in ["scatter", "hull"]:
            # For scatter and hull, join closest point to closest point
            for i, j in edge_index:
                vi, vj = points[clusts[i]], points[clusts[j]]
                i1, i2, _ = closest_pair(vi, vj, "recursive")
                edge_vertices.extend([vi[i1], vj[i2], [None, None, None]])

        else:
            # For cones, use the cone start points
            sts = []
            for trace in traces:
                start = [trace["x"][0], trace["y"][0], trace["z"][0]]
                sts.append(start)

            for i, j in edge_index:
                edge_vertices.extend([sts[i], sts[j], [None, None, None]])

        edge_vertices = np.vstack(edge_vertices)

    # Initialize the edge labels, if they are provided
    if edge_labels is not None:
        edge_labels = np.repeat(edge_labels, 3)

    # Add the edge trace
    traces += scatter_points(
        edge_vertices,
        color=edge_labels,
        line=line,
        linewidth=linewidth,
        mode="lines",
        name=edge_name,
    )

    # Return
    return traces


def network_schematic(
    clusts,
    edge_index,
    clust_labels,
    edge_labels=None,
    color=None,
    name=None,
    linewidth=2,
    **kwargs,
):
    """Network 2D schematic representation.

    This is to be used exclusevely with bipartite graphs where the nodes
    are either classified as primary or secondaries under clust_labels and
    connections only exist between primaries and secondaries.

    Parameters
    ----------
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    edge_index : np.ndarray
        (E, 2) List of connections between clusters
    clust_labels : np.ndarray
        (C) Whether a cluster is a primary or a secondary
    edge_labels : np.ndarray, optional
        (E) List of edge labels
    linewidth : float, default 2
        Width of the edge lines
    color : Union[str, np.ndarray], optional
        Color of clusters or (C) list of color of clusters
    name : str, optional
        Name of the network
    linewidth : float, default 2
        Width of the edge lines
    **kwargs : dict, optional
        List of additional arguments to pass to plotly

    Returns
    -------
    List[plotly.graph_objs.Scatter]
        Node and edge traces in the same list
    """
    # Check that color is not passed directly, ambiguous for a network
    assert (
        color is None
    ), "Use `clust_labels` instead of `color` to specify node colors."

    # Define the node size on the bases of the cluster size
    counts = np.array([len(c) for c in clusts])
    node_sizes = np.sqrt(counts)

    # Set the prefix to add to the trace names
    prefix = f"{name}" if name is not None else "Graph"
    node_name = name if edge_index is None else f"{prefix} nodes"
    edge_name = f"{prefix} edges"

    # Check that the labels are binary (0 or 1)
    assert len(clust_labels) == len(
        clusts
    ), "Must provide a primary label for each cluster."
    assert np.all(
        (clust_labels == 0) | (clust_labels == 1)
    ), "All cluster labels should be 0 or 1."

    # Define the hovertext attribute
    num_clusts = len(clusts)
    node_labels = []
    for i in range(num_clusts):
        node_labels.append(f"Cluster ID: {i:d}")
        node_labels[i] += f"<br>Primary: {clust_labels[i]:0.0f}"
        node_labels[i] += f"<br>Size: {counts[i]:d}"

    # Define the positions (primaries on the left, secondaries on the right)
    pos = np.array([[i, l] for i, l in enumerate(clust_labels)])

    # Define the trace associated with the graph nodes
    node_trace = scatter_points(
        pos,
        color=clust_labels,
        hovertext=node_labels,
        hoverinfo=["x", "y", "text"],
        markersize=node_sizes,
        name=node_name,
        **kwargs,
    )

    # Define the trace associated with the graph edges
    edge_vertices = np.empty((0, 2), dtype=pos.dtype)
    if len(edge_index):
        edge_vertices = []
        for i, j in edge_index:
            edge_vertices.extend([pos[i], pos[j], [None, None]])

        edge_vertices = np.vstack(edge_vertices)

    # Initialize the edge labels, if they are provided
    if edge_labels is not None:
        edge_labels = np.repeat(edge_labels, 3)

    # Add the edge trace
    edge_trace = []
    edge_trace = scatter_points(
        edge_vertices,
        color=edge_labels,
        linewidth=linewidth,
        mode="lines",
        name=edge_name,
    )

    return node_trace + edge_trace



"""Module which defines a very generic multi-layer perceptron with
fully configurable parameters to be used elsewhere.
"""

from typing import List, Union

import numpy as np
import torch
from torch import nn

from .act_norm import act_factory, norm_factory

__all__ = ["MLP"]


class MLP(nn.Module):
    """Generic multi-layer perceptron to be used as a feature extractor."""

    name = "mlp"

    def __init__(self, in_channels, depth, width, activation, normalization):
        """Initialize the MLP.

        Parameters
        ----------
        in_channels : int
            Number of input features
        depth : int
            Number of hidden layers
        width : Union[int, List[int]]
            Number of neurons in each hidden layer
        activation : Union[str, dict]
            Activation function configuration
        normalization : Union[str, dict]
            Normalization function configuration
        """
        # Initialize the parent class
        super().__init__()

        # Store the attribtues
        self.in_channels = in_channels
        self.depth = depth
        self.act_cfg = activation
        self.norm_cfg = normalization

        # Process the width
        if isinstance(width, int):
            self.width = [width] * depth
        else:
            assert len(width) == depth, (
                "If provided as an array, the `width` must be given "
                "once for each hidden layer (specified in `depth`)"
            )
            self.width = width

        self.feature_size = self.width[-1]

        # Initialize the model
        self.model = nn.Sequential()
        num_feats = in_channels
        for i in range(depth):
            # Add a layer of hidden neurons
            self.model.append(nn.Linear(num_feats, self.width[i]))
            self.model.append(norm_factory(normalization, self.width[i]))
            self.model.append(act_factory(activation))

            num_feats = self.width[i]

    def forward(self, x):
        """Pass a tensor of features through the MLP.

        Parameters
        ----------
        x : torch.Tensor
            (N, F) Tensor of features

        Paramters
        ---------
        torch.Tensor
            (N, W) Updated tensor of features
        """
        return self.model(x)



"""Class and methods to convert the output of Graph-SPICE into a set of
pixel cluster assignments using a graph.
"""

import sys
from collections import defaultdict
from functools import partial
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import torch
from torch_cluster import knn_graph, radius_graph

from spine.data import IndexBatch, ObjectList, TensorBatch
from spine.utils.enums import enum_factory
from spine.utils.globals import CLUST_COL, SHAPE_COL
from spine.utils.gnn.cluster import form_clusters
from spine.utils.metrics import ari, eff, pur

from .ccc import ConnectedComponentClusterer

__all__ = ["ClusterGraphConstructor"]


class ClusterGraphConstructor:
    """Manager class for handling per-batch, per-semantic type graph
    construction and node predictions in Graph-SPICE clustering.
    """

    def __init__(
        self,
        graph,
        shapes,
        edge_threshold,
        kernel_fn=None,
        min_size=0,
        invert=True,
        label_edges=False,
        target_col=CLUST_COL,
        training=False,
        orphan=None,
    ):
        """Initialize the cluster graph constructor.

        Parameters
        ----------
        graph : dict
            Graph construction configuration dictionary
        shapes : List[str]
            List of shape names to construct clusters for
        edge_threshold : float
            Edge score below which it is disconnected (or above which it is,
            if the `inverted` parameter is turned on
        kernel_fn : callable, optional
            Kernel function computing edge scores from edge features
        min_size : int, default 0
            Minimum number of points below which pixels are considered orphans
            to be merged into touching larger clusters
        invert : bool, default True
            Invert the edge scores so that 0 is on an 1 is off
        label_edges : bool, default False
            If `True`, use cluster labels to label the edges as on or off
        target_col : int, default CLUST_COL
            Index of the column which specifies the label cluster ID for each point
        training : bool, default False
            If `True`, this constructor is being used at train time
        orphan : dict, optional
            Orphan clustering configuration dictionary

        Raises
        ------
        ValueError
            If the graph type is not supported.
        """
        # Parse the set of shapes to cluster
        self.shapes = enum_factory("shape", shapes)

        # Store other basic properties
        self.threshold = edge_threshold
        self.min_size = min_size
        self.invert = invert
        self.label_edges = label_edges
        self.kernel_fn = kernel_fn
        self.target_col = target_col

        # Partially instantiate the graph constructor functions
        assert "name" in graph, "Must provide the graph constructor function name."

        name = graph.pop("name")
        if name == "knn":
            self.graph_fn = partial(knn_graph, **graph)
        elif name == "radius":
            self.graph_fn = partial(radius_graph, **graph)
        else:
            raise ValueError(
                f"Requested graph construction mode ('{name}') is not "
                "recognized. Must be one of 'knn' or 'radius'"
            )

        # Initialize the cluster assignment class
        self.ccc = ConnectedComponentClusterer(min_size, orphan)

    def __call__(self, coords, features, seg_label, clust_label=None):
        """Constructs graphs for all the entries in a batch, one per shape.

        Parameters
        ----------
        coords : TensorBatch
            (N, 3) Point coordinates
        features : TensorBatch
            (N, N_f) Set of graph embeddings
        seg_label : TensorBatch
            (N, 1 + D + 1) Tensor of segmentation labels
            - 1 is the segmentation label
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
            - N_c is is the number of cluster labels
        """
        # If edge labeling is required, make sure clust_label is provided
        assert (
            not self.label_edges or clust_label is not None
        ), "If edge labels are to be produced, must provide `clust_label`."

        # Loop over the unique batch indices, build a list of graphs for each
        graph = defaultdict(list)
        edge_offset = 0
        edge_counts, edge_offsets = [], []
        for b in range(coords.batch_size):
            # Build graphs (one per semantic type)
            clust_label_b = clust_label[b] if clust_label is not None else None
            graphs_b, edge_count = self.build_graph(
                coords[b], features[b], seg_label[b], clust_label_b
            )

            # Append the output
            edge_counts.append(edge_count)
            edge_offsets.append(edge_offset)
            for key, value in graphs_b.items():
                if key.endswith("clusts"):
                    is_edge = key.startswith("edge")
                    offset = edge_offset if is_edge else coords.edges[b]
                    for c in value:
                        graph[key].append(offset + c)

                else:
                    graph[key].append(value)

            # Increment the offset
            edge_offset += edge_count

        # Concatenate the graph attributes together
        is_tensor = isinstance(coords.tensor, torch.Tensor)
        cat = torch.cat if is_tensor else np.concatenate
        for key, value in graph.items():
            if key.endswith("clusts"):
                # Turn indexes into index batches
                counts = [len(self.shapes)] * coords.batch_size
                single_counts = [len(c) for c in value]
                is_edge = key.startswith("edge")
                offsets = edge_offsets if is_edge else coords.edges[:-1]
                graph[key] = IndexBatch(value, offsets, counts, single_counts)

            else:
                # Turn edge index/attributes into tensor batches
                value = cat(value)
                graph[key] = TensorBatch(value, edge_counts)

        # Add the input node information to the graph
        graph["node_coords"] = coords
        graph["node_features"] = features
        graph["node_shapes"] = TensorBatch(
            seg_label.tensor[:, SHAPE_COL], seg_label.counts
        )

        return graph

    def build_graph(self, coords, features, seg_label, clust_label=None):
        """Construct a graph for a single batch id and semantic class that
        will be used for connected components clustering.

        Parameters
        ----------
        coords : torch.Tensor
            (N, 3) Tensor of point coordinates
        features : torch.Tensor
            (N, N_f) Graph embedding features to be used for edge prediction
        seg_clusts : List[List[int]]
            (S) One pixel index per semantic type
        seg_label : torch.Tensor
            (N, 1 + D + 1) Tensor of segmentation labels
            - 1 is the segmentation label
        clust_label : torch.Tensor, optional
            (N, 1 + D + N_c) Tensor of cluster labels
            - N_c is is the number of cluster labels

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary of graph properties
        """
        # Loop over the semantic types, build a graph for each
        graph = defaultdict(list)
        edge_count = 0
        for s in self.shapes:
            # Get the index of points which belong to this class
            seg_index = torch.where(seg_label[:, SHAPE_COL] == s)[0]
            graph["node_clusts"].append(seg_index)

            # If there are no points, append empty, proceed
            if not len(seg_index):
                graph["edge_clusts"].append(
                    torch.empty(0, dtype=torch.long, device=coords.device)
                )
                graph["edge_index"].append(
                    torch.empty((0, 2), dtype=torch.long, device=coords.device)
                )
                graph["edge_shape"].append(
                    torch.empty(0, dtype=torch.long, device=coords.device)
                )
                graph["edge_attr"].append(
                    torch.empty(0, dtype=features.dtype, device=features.device)
                )
                if self.label_edges:
                    graph["edge_label"].append(
                        torch.empty(0, dtype=torch.long, device=coords.device)
                    )
                continue

            # Make the graph edge index
            edge_index = self.graph_fn(coords[seg_index])
            graph["edge_clusts"].append(
                edge_count + torch.arange(edge_index.shape[1], device=coords.device)
            )
            edge_count += edge_index.shape[1]

            # Produce edge predictions
            features_s = features[seg_index]
            edge_attr = self.kernel_fn(
                features_s[edge_index[0]], features_s[edge_index[1]]
            )

            # Append
            graph["edge_index"].append(edge_index.T)
            graph["edge_shape"].append(
                torch.full(
                    (edge_index.shape[1],), s, dtype=torch.long, device=coords.device
                )
            )
            graph["edge_attr"].append(edge_attr.flatten())

            if self.label_edges:
                node_label = clust_label[seg_index, self.target_col]
                edge_label = node_label[edge_index[0]] == node_label[edge_index[1]]
                graph["edge_label"].append(edge_label.long())

        # Concatenate the graph attributes
        for key, value in graph.items():
            if key != "node_clusts" and key != "edge_clusts":
                graph[key] = torch.cat(value)

        # Convert edge logits to sigmoid scores
        graph["edge_prob"] = torch.sigmoid(graph["edge_attr"])

        return graph, edge_count

    def fit_predict(self, graph, edge_mode="edge_pred", threshold=None, min_size=None):
        """Perform connected components clustering on a batch.

        Parameters
        ----------
        graph : dict
            Dictionary of graph attributes organized by batch and shape
        edge_mode : str, default 'edge_pred'
            Attribute of the graph used to get the edge status
        threshold : float, optional
            Override the edge score threshold set in the configuration
        min_size : int, optional
            Override the minimum cluster size set in the configuration

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            Node assignments
        """
        # No gradients through this prediction
        with torch.no_grad():
            # Assign edge predictions based on the edge scores
            threshold = threshold if threshold is not None else self.threshold
            if self.invert:
                edge_pred = (graph["edge_prob"].tensor <= threshold).long()
            else:
                edge_pred = (graph["edge_prob"].tensor >= threshold).long()

            graph["edge_pred"] = TensorBatch(edge_pred, graph["edge_prob"].counts)

            # Assign each node to a cluster
            node_pred = self.ccc(
                graph["node_coords"],
                graph["edge_index"],
                graph[edge_mode],
                graph["node_clusts"],
                graph["edge_clusts"],
            )

            graph["node_pred"] = node_pred

            # Loop over entries in the batch, build fragments
            node_clusts = graph["node_clusts"]
            clusts, counts, single_counts, shapes = [], [], [], []
            for b in range(node_pred.batch_size):
                # Loop over shapes in the entry
                counts_b = 0
                for s, shape in enumerate(self.shapes):
                    # Get the list of clusters for this (entry, shape) pair
                    index_b_s = node_clusts[b][s]
                    clusts_b_s, counts_b_s = form_clusters(
                        node_pred[b][index_b_s, None], column=0
                    )

                    # Offset the cluster indexes appropriately, append
                    for i, c in enumerate(clusts_b_s):
                        clusts_b_s[i] = node_pred.edges[b] + index_b_s[c]

                    # Append
                    clusts.extend(clusts_b_s)
                    single_counts.extend(counts_b_s)
                    shapes.append(shape * np.ones(len(clusts_b_s), dtype=int))
                    counts_b += len(clusts_b_s)

                counts.append(counts_b)

            # Make an IndexBatch out of the list
            clusts = IndexBatch(clusts, node_pred.edges[:-1], counts, single_counts)
            clust_shapes = TensorBatch(np.concatenate(shapes), counts)

            # Return
            return clusts, clust_shapes

    def evaluate(self, graph, mean=False):
        """Evaluate the clustering accuracy of a graph.

        Parameters
        ----------
        graph : dict
            Dictionary of graph attributes organized by batch and shape
        mean : bool, default False
            If `True`, returns the batch-averaged metric values

        Returns
        -------
        dict
            Dictionary of accuracy metrics
        """
        # No gradients through this evaluation
        result = defaultdict(list)
        metrics = {"ari": ari, "purity": pur, "efficiency": eff}
        with torch.no_grad():
            # Loop over the batches
            for b in range(batch_size):
                # Get the node predictions and labels
                node_label_b = graph["node_label"][b]
                node_pred_b = graph["node_pred"][b]

                # Compute shape-agnostic metrics
                for m, metric in metrics.items():
                    result[m].append(metric(node_pred_b, node_label_b))

                # Loop over the semantic types
                for s, shape in self.shapes:
                    # Narrow down the predictions and labels to this shape
                    node_index = graph["node_clusts"][b][s]
                    node_label_b_s = node_label_b[node_index]
                    node_pred_b_s = node_pred_b[node_index]

                    # If there are no points of this type, append default values
                    if not len(node_index):
                        for m in metrics:
                            result[f"{m}_{shape}"].append(1.0)

                    # Otherwise, compute the metrics
                    for m, metric in metrics.items():
                        result[f"{m}_{shape}"].append(
                            metric(node_pred_b_s, node_label_b_s)
                        )

        # Compute batch averaged metrics, return
        if mean:
            for key, value in result.items():
                result[key] = np.mean(value)

        return result

    @staticmethod
    def get_entry(graph, batch_id, semantic_id):
        """Narrow down the graph to one specific (batch_id, shape) pair.

        Parameters
        ----------
        graph : dict
            Dictionary of graph attributes organized by batch and shape
        batch_id : int
            Batch index
        semantic_id : int
            Semantic type

        Returns
        -------
        dict
            Dictionary of graph attributes for one (batch_id, shape) pair
        """
        # Loop over graph keys, narrow down wherever relevant
        single_graph = {}
        node_index = graph["node_clusts"][batch_id][semantic_id]
        edge_index = graph["edge_clusts"][batch_id][semantic_id]
        for key, value in graph.items():
            if key.endswith("clusts"):
                continue
            elif key.startswith("node"):
                single_graph[key] = graph[key][b][node_index]
            elif key.startswith("edge"):
                single_graph[key] = graph[key][b][edge_index]
            else:
                raise KeyError(f"Graph key not recognized: {key}")

        return single_graph



"""Particle topology module."""

import numpy as np
from scipy.stats import pearsonr

from spine.math.decomposition import PCA
from spine.post.base import PostBase
from spine.utils.globals import ELEC_PID, PHOT_PID, SHOWR_SHP
from spine.utils.gnn.cluster import cluster_dedx, cluster_dedx_dir

__all__ = [
    "ParticleDEDXProcessor",
    "ParticleStartStraightnessProcessor",
    "ParticleSpreadProcessor",
]


class ParticleDEDXProcessor(PostBase):
    """Compute the dE/dx of the particle start by summing the energy depositions
    along the particle start and dividing by the total length of the start.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "local_dedx"

    # Aliases for the post-processor
    aliases = ("start_dedx", "end_dedx")

    # List of recognized dE/dx computation modes
    _modes = ("default", "direction")

    def __init__(
        self,
        radius=3.0,
        anchor=False,
        mode="direction",
        include_pids=(PHOT_PID, ELEC_PID),
        include_secondary=False,
    ):
        """Store the particle start dE/dx reconstruction parameters.

        Parameters
        ----------
        radius : float, default 3.0
            Radius around the start point to include in the dE/dx calculation.
        anchor : bool, default False
            If True, anchor the start point to the particle cluster. This
            is irrelevant for reconstruction particles (always anchored)
        mode : str, default 'direction'
            Method to use for dE/dx calculation.
        include_pids : list, default [0, 1]
            Particle species to compute the start dE/dx for
        include_secondary : bool, default False
            If `True`, computes the start dE/dx for secondary particles
        """
        # Initialize the parent class
        super().__init__("particle", "reco")

        # Check that the dE/dx mode is valid, store
        assert mode in self._modes, (
            f"dE/dx computation mode not recognized: {mode}. Should be "
            f"one of {self._modes}"
        )
        self.mode = mode

        # Store the dE/dx computation parameters
        self.radius = radius
        self.anchor = anchor
        self.include_pids = include_pids
        self.include_secondary = include_secondary

        # If the method involves the direction, must run the direction PP
        if mode == "direction":
            self.update_upstream("direction")

    def process(self, data):
        """Compute the start dE/dx for all particles in one entry.

        Parameters
        ----------
        data : dict
            Dictionaries of data products
        """
        # Loop over the reco particles
        for part in data["reco_particles"]:
            # If the particle is a secondary and they are to be skipped, do
            if not self.include_secondary and not part.is_primary:
                continue

            # If the particle PID is not in the required list, skip
            if part.pid not in self.include_pids:
                continue

            # Loop over the two sides of the particle
            for side in ("start", "end"):
                # Showers have no end points, skip
                if side == "end" and part.shape == SHOWR_SHP:
                    continue

                # Fetch the point
                ref_point = getattr(part, f"{side}_point")

                # Compute the particle end dE/dx
                if self.mode == "default":
                    # Use all depositions within a radius of the particle point
                    dedx = cluster_dedx(
                        part.points,
                        part.depositions,
                        ref_point,
                        max_dist=self.radius,
                        anchor=self.anchor,
                    )

                else:
                    # Use the particle direction estimate as a guide
                    ref_dir = getattr(part, f"{side}_dir") * (-1) ** (side == "end")
                    dedx = cluster_dedx_dir(
                        part.points,
                        part.depositions,
                        ref_point,
                        ref_dir,
                        max_dist=self.radius,
                        anchor=self.anchor,
                    )[0]

                # Store the dE/dx
                setattr(part, f"{side}_dedx", dedx)


class ParticleStartStraightnessProcessor(PostBase):
    """Compute the relative straightness of the particle start.

    The start straightness is evaluated by computing the covariance matrix of
    the points in a neighborhood around the start of a particle and by
    estimating which fraction of that variance is explained by the first
    principal component of the decomposition.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "start_straightness"

    def __init__(
        self,
        radius=3.0,
        n_components=3,
        include_pids=(PHOT_PID, ELEC_PID),
        include_secondary=False,
    ):
        """Store the particle start straightness reconstruction parameters.

        Parameters
        ----------
        radius : float, default 3.0
            Radius around the start point to include in the straightness calculation.
        n_components : int, default 3
            Number of components to compute the PCA for
        include_pids : list, default [0, 1]
            Particle species to compute the start dE/dx for
        include_secondary : bool, default False
            If `True`, computes the start dE/dx for secondary particles
        """
        # Initialize the parent class
        super().__init__("particle", "reco")

        # Check that the number of components make sense, initialize PCA
        assert n_components > 0 and n_components <= 3, (
            "The number of PCA components should be at least 1 and at most "
            "3 (the dimensionality of 3D points)."
        )
        self.n_components = n_components
        self.pca = PCA(n_components=n_components)

        # Store the dE/dx computation parameters
        self.radius = radius
        self.include_pids = include_pids
        self.include_secondary = include_secondary

    def process(self, data):
        """Compute the start straightness for all particles in one entry.

        Parameters
        ----------
        data : dict
            Dictionaries of data products
        """
        # Loop over the reco particles
        for part in data["reco_particles"]:
            # If the particle is a secondary and they are to be skipped, do
            if not self.include_secondary and not part.is_primary:
                continue

            # If the particle PID is not in the required list, skip
            if part.pid not in self.include_pids:
                continue

            # Compute the particle start straightness
            part.start_straightness = self.get_start_straightness(part)

    def get_start_straightness(self, particle):
        """Evaluates the straightness of the start of a particle by computing
        the PCA principal explained variance ratio.

        Parameters
        ----------
        particle : RecoParticle
            Reconstructed particle instance

        Returns
        -------
        float
            The first explained variance ratio of the PCA of the shower start
        """
        # Restrict the points to those around the start of the particle
        dists = np.linalg.norm(particle.points - particle.start_point, axis=1)
        points = particle.points[dists < self.radius]

        # Compute the explained variance ratio of the principal component
        if len(points) < self.n_components:
            return -1.0
        else:
            _, explained_variance = self.pca.fit(points)
            total_variance = np.sum(explained_variance)
            return explained_variance[0] / total_variance if total_variance > 0 else 0.0


class ParticleSpreadProcessor(PostBase):
    """Compute the directional and axial spreads of a particle.

    The directional spread is obtained by computing the mean direction and the
    weighted average cosine distance with respect to the mean direction.

    The axial spread is obtained by computing the pearson R correlation
    coefficient between the distance of the shower points from the startpoint
    along the shower axis and the perpendicular distance from the shower axis.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "particle_spread"

    # List of recognized start reference modes for spread evaluation
    _start_modes = ("vertex", "start_point")

    # Set of post-processors which must be run before this one is
    _upstream = ("direction",)

    def __init__(
        self,
        start_mode="vertex",
        length_scale=14.0,
        use_start_dir=False,
        include_pids=(PHOT_PID, ELEC_PID),
        include_secondary=False,
    ):
        """Store the particle spread reconstruction parameters.

        Parameters
        ----------
        start_mode : str, default 'vertex'
            Point used as a start point to evaluate the spread (can be one
            of 'vertex' or 'start_point')
        length_scale : float, default 14.0
            Length scale used to weight the directional spread estimator
        use_start_dir : bool, default False
            If `True`, use the direction estimate of the particle to compute
            the axial spread. If `False`, the direction is estimated using PCA
        include_pids : list, default [0, 1]
            Particle species to compute the start dE/dx for
        include_secondary : bool, default False
            If `True`, computes the start dE/dx for secondary particles
        """
        # Initialize the parent class
        super().__init__("interaction", "reco")

        # Check that the start point mode is valid, store
        assert start_mode in self._start_modes, (
            f"Spread start computation mode not recognized: {start_mode}. "
            f"Should be one of {self._start_modes}"
        )
        self.start_mode = start_mode

        # Store the spread computation parameters
        self.length_scale = length_scale
        self.use_start_dir = use_start_dir
        self.include_pids = include_pids
        self.include_secondary = include_secondary

        # If needed, initialize PCA
        if not use_start_dir:
            self.pca = PCA(n_components=3)

    def process(self, data):
        """Compute the shower spread for all particles in one entry.

        Parameters
        ----------
        data : dict
            Dictionaries of data products
        """
        # Loop over the reco interactions
        for inter in data["reco_interactions"]:
            # Loop over the particles that make up the interaction
            for part in inter.particles:
                # If the particle is a secondary and they are to be skipped, do
                if not self.include_secondary and not part.is_primary:
                    continue

                # If the particle PID is not in the required list, skip
                if part.pid not in self.include_pids:
                    continue

                # Fetch the appropriate reference point
                if self.start_mode == "vertex":
                    ref_point = inter.vertex
                else:
                    ref_point = part.start_point

                # Compute the spreads
                part.directional_spread = self.get_dir_spread(part.points, ref_point)
                part.axial_spread = self.get_axial_spread(
                    part.points, part.start_dir, ref_point
                )

    def get_dir_spread(self, points, ref_point):
        """Compute the directional spread of the particle by computing the mean
        direction and the weighted average cosine distance with respect to it.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Particle point coordinates
        ref_point : np.ndarray
            (3) Reference point

        Returns
        -------
        float
            Directional spread of the particle
        """
        # Compute the distance from the reference point to all particle points
        dists = np.linalg.norm(points - ref_point, axis=1)

        # Restrict points to those that do not coincide with the reference point
        index = np.where(dists > 0)[0]
        if len(index) == 0:
            return -1.0
        points, dists = points[index], dists[index]

        # Compute the exponential-weighted directional spread
        directions = (points - ref_point) / dists.reshape(-1, 1)
        weights = np.maximum(np.exp(-dists / self.length_scale), 1e-6)
        mean_direction = np.average(directions, weights=weights, axis=0)
        mean_direction /= np.linalg.norm(mean_direction)
        cosine = 1 - np.dot(directions, mean_direction)
        spread = np.average(cosine, weights=weights)

        return spread

    def get_axial_spread(self, points, start_dir, ref_point):
        """Compute the axial spread of a particle by evaluating the pearson R
        correlation coefficient between the distance of the particle points from
        the start point along the particle axis and the perpendicular distance
        from the particle axis.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Particle point coordinates
        start_dir : np.ndarray
            (3) Start direction of the particle
        ref_point : np.ndarray
            (3) Reference point

        Returns
        -------
        float
            Axial spread of the particle
        """
        # If there are less than 3 points, one cannot compute this quantity
        if len(points) < 3:
            return -np.inf

        # Fetch the particle direction
        if self.use_start_dir:
            v0 = start_dir
        else:
            components, _ = self.pca.fit(points)
            v0 = components[0]  # First principal component
            if np.dot(start_dir, v0) < 0.0:
                v0 *= -1

        # Compute the axial spread
        dists = np.linalg.norm(points - ref_point, axis=1)
        v = (ref_point - points) - (
            np.sum((ref_point - points) * v0, axis=1, keepdims=True)
            * np.broadcast_to(v0, points.shape)
        )
        perps = np.linalg.norm(v, axis=1)

        return pearsonr(dists, perps)[0]



"""Match objects and their label counterparts and vice versa."""

from dataclasses import dataclass

import numba as nb
import numpy as np
from numba import typed

import spine.utils.match
from spine.post.base import PostBase

__all__ = ["MatchProcessor"]


class MatchProcessor(PostBase):
    """Does the matching between reconstructed and true objects."""

    # Name of the post-processor (as specified in the configuration)
    name = "match"

    def __init__(
        self,
        fragment=None,
        particle=None,
        interaction=None,
        truth_point_mode="points",
        **kwargs,
    ):
        """Initializes the matching post-processor.

        Parameters
        ----------
        fragment: Union[bool, dict], optional
            Matching flag or configuration for fragments
        particle: Union[bool, dict], optional
            Matching flag or configuration for particles
        interaction: Union[bool, dict], optional
            Matching flag or configuration for interactions
        truth_point_mode : str, default 'points'
            Type of truth points to use to do the matching
        **kwargs : dict, optional
            Matching parameters shared between all matching processes
        """
        # Initialize the necessary matchers
        configs = {
            "fragment": fragment,
            "particle": particle,
            "interaction": interaction,
        }
        keys = {}
        self.matchers = {}
        for key, cfg in configs.items():
            if cfg is not None and cfg is not False:
                # Initialize the matcher
                if isinstance(cfg, bool):
                    cfg = {}
                self.matchers[key] = self.Matcher(**cfg, **kwargs)

                # If any matcher includes ghost points, must load meta
                if self.matchers[key].ghost:
                    keys["meta"] = True

        assert len(
            self.matchers
        ), "Must specify one of 'fragment', 'particle' or 'interaction'."

        # Update the set of keys necessary for this post-processor
        self.update_keys(keys)

        # Initialize the parent class
        super().__init__(list(self.matchers.keys()), "both", truth_point_mode)

    @dataclass
    class Matcher:
        """Simple data class to store matching methods per object.

        Attributes
        ----------
        fn : function
            Function which computes overlaps between pairs of objects
        match_mode : str, defualt 'both'
            Matching mode. One of 'reco_to_truth', 'truth_to_reco' or 'both'
        overlap_mode : str, default 'iou'
            Overlap estimatation method. One of 'count', 'iou', 'dice', 'chamfer'
        min_overlap : float, default 0.
            Overlap value above which a pair is considered a match
        weight_overlap : bool, default False
            Whether to weight the overlap metric
        ghost : bool, default False
            Whether a deghosting process was applied (in which case the indexes
            of the reco and the truth particles do not align)
        """

        fn: object = None
        match_mode: str = "both"
        overlap_mode: str = "iou"
        min_overlap: float = 0.0
        weight_overlap: bool = False
        ghost: bool = False

        # Valid match modes
        _match_modes = ["reco_to_truth", "truth_to_reco", "both", "all"]

        # Valid overlap modes
        _overlap_modes = ["count", "iou", "dice", "chamfer"]

        def __post_init__(self):
            """Check that the values provided are valid."""
            # Check match mode
            assert self.match_mode in self._match_modes, (
                f"Invalid matching mode: {self.match_mode}. Must be one "
                f"of {self._match_modes}."
            )

            # Check the overlap mode
            assert self.overlap_mode in self._overlap_modes, (
                f"Invalid overlap computation mode: {self.overlap_mode}. "
                f"Must be one of {self._overlap_modes}."
            )

            # Check that the overlap mode and weighting are compatible
            assert not self.weight_overlap or self.overlap_mode in [
                "iou",
                "dice",
            ], "Only IoU and Dice-based overlap functions can be weighted."

            # Initialize the match overlap function
            prefix = "overlap" if not self.weight_overlap else "overlap_weighted"
            self.fn = getattr(spine.utils.match, f"{prefix}_{self.overlap_mode}")

    def process(self, data):
        """Match all the requested objects in one entry.

        Parameters
        ----------
        data: dict
            Dictionary of data products
        """
        # Loop over the matchers
        result = {}
        for name, matcher in self.matchers.items():
            # Fetch the required data products
            reco_objs = data[f"reco_{name}s"]
            truth_objs = data[f"truth_{name}s"]

            # Fetch the metadata, if needed
            meta = None
            if matcher.ghost:
                meta = data["meta"]

            # Pass it to the individual processor
            res_one = self.process_single(reco_objs, truth_objs, matcher, name, meta)
            result.update(**res_one)

        return result

    def process_single(self, reco_objs, truth_objs, matcher, name, meta=None):
        """Match all the requested objects in a single category.

        Parameters
        ----------
        reco_objs : List[object]
            List of reconstructed objects
        truth_objs : List[object]
            List of truth objects
        matcher : MatchProcessor.Matcher
            Matching method and function
        name : str
            Object type name
        meta : Meta, optional
            Metadata information to convert position to index
        """
        # Convert the object list into an index/coordinate list
        if matcher.overlap_mode != "chamfer":
            # For overlap matches, use pixel indexes (faster)
            if not matcher.ghost or self.truth_point_mode == "points_adapt":
                # The indexes of reco and truth point to the same point set
                reco_input = typed.List.empty_list(nb.int64[:])
                for p in reco_objs:
                    reco_input.append(self.get_index(p))
                truth_input = typed.List.empty_list(nb.int64[:])
                for p in truth_objs:
                    truth_input.append(self.get_index(p))

            else:
                # The indexes of reco and truth point to different point sets.
                # In this case, convert the positions to indexes
                assert meta is not None, (
                    "When using separate point sets for reco and truth objects, "
                    "must provide the metadata object to convert positions to "
                    "indexes."
                )
                reco_input = []
                for p in reco_objs:
                    coords = self.get_points(p)
                    if p.units != "px":
                        coords = meta.to_px(coords, floor=True)
                    reco_input.append(meta.index(coords))

                truth_input = []
                for p in truth_objs:
                    coords = self.get_points(p)
                    if p.units != "px":
                        coords = meta.to_px(coords, floor=True)
                    truth_input.append(meta.index(coords))

        else:
            # For the chamfer distance, simply use the point positions
            reco_input = typed.List.empty_list(nb.float32[:, :])
            for p in reco_objs:
                reco_input.append(self.get_points(p))
            truth_input = typed.List.empty_list(nb.float32[:, :])
            for p in truth_objs:
                truth_input.append(self.get_points(p))

        # Pass lists to the matching function to compute overlaps
        if len(reco_input) and len(truth_input):
            ovl_matrix = matcher.fn(reco_input, truth_input)
        else:
            ovl_matrix = np.empty((len(reco_input), len(truth_input)))

        # Make the overlap selection cut, if requested
        if matcher.overlap_mode != "chamfer":
            ovl_valid = ovl_matrix > matcher.min_overlap
        else:
            ovl_valid = ovl_matrix < matcher.min_overlap

        # Produce matches
        result = {}
        if matcher.match_mode != "truth_to_reco":
            pairs, overlaps = self.generate_matches(
                reco_objs, truth_objs, ovl_matrix, ovl_valid
            )
            result[f"{name}_matches_r2t"] = pairs
            result[f"{name}_matches_r2t_overlap"] = overlaps

        if matcher.match_mode != "reco_to_truth":
            pairs, overlaps = self.generate_matches(
                truth_objs, reco_objs, ovl_matrix.T, ovl_valid.T
            )
            result[f"{name}_matches_t2r"] = pairs
            result[f"{name}_matches_t2r_overlap"] = overlaps

        return result

    @staticmethod
    def generate_matches(source_objs, target_objs, ovl_matrix, ovl_valid):
        """Generate pairs for a srt of sources and targets.

        Parameters
        ----------
        source_objs : List[object]
            (N) List of source objects
        target_objs : List[object]
            (M) List of truth objects
        ovl_matrix : np.ndarray
            (N, M) Matrix of overlap values
        ovl_valid : np.ndarray
            (N, M) Matrix of overlap validity

        Returns
        -------
        pairs : List[tuple]
            (N) List of (source, target) matched pairs (best match only)
        overlaps : List[float]
            (N) List of overlap between each source and the best matched target
        """
        # Build the matches based on the threshold
        pairs, pair_overlaps = [], []
        for i, s in enumerate(source_objs):
            # Get the list of valid matches
            match_idxs = np.where(ovl_valid[i])[0]
            if not len(match_idxs):
                # If there are no matches, fill dummy values
                s.is_matched = False
                s.match_ids = np.empty(0, dtype=np.int64)
                s.match_overlaps = np.empty(0, dtype=np.float32)

                pairs.append((s, None))
                pair_overlaps.append(-1.0)

            else:
                # If there are matches, order them by decreasing overlap
                overlaps = ovl_matrix[i, match_idxs]
                perm = np.argsort(overlaps)[::-1]
                s.is_matched = True
                s.match_ids = match_idxs[perm]
                s.match_overlaps = overlaps[perm]

                best_idx = s.match_ids[0]
                pairs.append((s, target_objs[best_idx]))
                pair_overlaps.append(s.match_overlaps[0])

        # Fill the match lists
        return pairs, pair_overlaps



"""Utility functions and tools used across the SPINE package.

This module contains reusable utility functions organized into several categories:

**Core Utilities:**
- `conditional`: Conditional imports for optional dependencies (ROOT, LArCV, MinkowskiEngine, PyTorch)
- `config`: Configuration file parsing and validation
- `factory`: Generic factory pattern implementations
- `logger`: Logging utilities and configuration
- `globals`: Global constants and enumerations

**Decorators:**
- `docstring`: Docstring inheritance utilities for class hierarchies
- `jit`: JIT compilation decorators using Numba for performance optimization

**Mathematical & Scientific:**
- `cluster`: CNN clustering algorithms and utilities
- `gnn`: Graph Neural Network utilities (network operations, clustering, voxel processing)
- `particles`: Particle physics calculations and utilities
- `pid`: Particle identification algorithms
- `tracking`: Track reconstruction algorithms
- `vertex`: Vertex reconstruction utilities

**PyTorch Utilities:**
- `torch`: PyTorch utilities organized into submodules:
  - `torch.runtime`: Tensor operations, CUDA memory management, distributed operations
  - `torch.devices`: CUDA device configuration and distributed training setup
  - `torch.training`: Optimizer and learning rate scheduler factories
  - `torch.scripts`: Utility scripts and PyTorch function extensions (e.g., cdist_fast)
  - `torch.adabound`: Custom AdaBound and AdaBoundW optimizer implementations

**Hardware & Performance:**
- `energy_loss`: Energy loss calculations and CSDA tables
- `optical`: Optical detector simulation and analysis
- `calib`: Calibration utilities and corrections

**Analysis Tools:**
- `ppn`: Point Proposal Network utilities
- `match`: Truth matching algorithms
- `metrics`: Evaluation metrics and statistics
- `inference`: Model inference utilities
- `stopwatch`: Performance timing utilities

**Data Processing:**
- `unwrap`: Data unwrapping and processing utilities
- `weighting`: Sample weighting and balancing

Most utilities are designed to be framework-agnostic where possible, with PyTorch-specific
functionality clearly separated and conditionally imported.
"""



import os
import re
import time
from pprint import pprint

import numpy as np
import pandas as pd
import torch
import yaml

from spine.driver import Driver
from spine.io.factories import loader_factory
from spine.math.metrics import adjusted_rand_score as ari
from spine.utils.metrics import *


def get_optimal_parameters(fname, metric="SBD"):
    """
    Retrives optimal seediness and probability thresholds for dense clustering.

    INPUTS:
        - fname: post-processing output filename (*.csv file)
        - metric: name of performance metric for which to optimize values on.

    RETURNS:
        - s_thresholds: dict of { semantic class : optimal seediness threshold } pairs
        - p_thresholds: dict of { semantic class : optimal pvalue threshold } pairs
        - max_values: maximum value of metric using optimal thresholding values.

    """
    data = pd.read_csv(fname)
    s_thresholds = {}
    p_thresholds = {}
    max_values = {}
    #     max_aris = {}
    for c in labels:
        means = (
            data.query("Class == {}".format(c))
            .groupby(["seed_threshold", "prob_threshold"])
            .mean()
        )
        if metric == "f1":
            metric_df = 2 * means["SBD"] * means["ARI"] / (means["ARI"] + means["SBD"])
        elif metric == "avg":
            metric_df = (means["SBD"] + means["ARI"]) / 2
        else:
            metric_df = means[metric]
        s_thresholds[c] = metric_df.idxmax()[0]
        p_thresholds[c] = metric_df.idxmax()[1]
        max_values[c] = metric_df.max()
    #             max_aris[c] = means['ARI'].max
    return s_thresholds, p_thresholds, max_values


def make_inference_cfg(train_cfg, gpu=1, snapshot=None, batch_size=1, model_path=None):
    from spine.main_funcs import process_config

    cfg = yaml.load(open(train_cfg, "r"), Loader=yaml.Loader)
    process_config(cfg)
    inference_cfg = cfg.copy()

    # Change batch size to 1 since no need for batching during validation
    inference_cfg["io"]["batch_size"] = batch_size
    inference_cfg["io"].pop("minibatch_size", None)
    inference_cfg["trainval"]["gpus"] = str(gpu)
    inference_cfg["trainval"]["train"] = False

    # Analysis keys for clustering
    inference_cfg["model"]["analysis_keys"] = {
        "segmentation": 0,
        "clustering": 1,
    }

    # Get latest model path if checkpoint not provided.
    if model_path is None:
        model_path = inference_cfg["trainval"]["model_path"]
    else:
        inference_cfg["trainval"]["model_path"] = model_path
    if snapshot is None:
        checkpoints = [
            int(re.findall("snapshot-([0-9]+).ckpt", f)[0])
            for f in os.listdir(re.sub(r"snapshot-([0-9]+).ckpt", "", model_path))
            if "snapshot" in f
        ]
        print(checkpoints)
        latest_ckpt = max(checkpoints)
        model_path = re.sub(
            r"snapshot-([0-9]+)", "snapshot-{}".format(str(latest_ckpt)), model_path
        )
    else:
        model_path = re.sub(
            r"snapshot-([0-9]+)", "snapshot-{}".format(snapshot), model_path
        )
    inference_cfg["trainval"]["model_path"] = model_path
    pprint(inference_cfg)
    process_config(inference_cfg)
    return inference_cfg


def gaussian_kernel(centroid, sigma):
    def f(x):
        dists = np.sum(np.power(x - centroid, 2), axis=1, keepdims=False)
        probs = np.exp(-dists / (2.0 * sigma**2))
        return probs

    return f


def gaussian_kernel_cuda(centroid, sigma):
    def f(x):
        dists = torch.sum(torch.pow(x - centroid, 2), dim=1)
        probs = torch.exp(-dists / (2.0 * sigma**2))
        return probs

    return f


def ellipsoidal_kernel(centroid, sigma):
    def f(x):
        dists = np.power(x - centroid, 2) / (2.0 * sigma**2)
        probs = np.exp(-np.sum(-dists, axis=1, keepdims=False))
        return probs

    return f


def find_cluster_means(features, labels):
    r"""
    For a given image, compute the centroids \mu_c for each
    cluster label in the embedding space.

    INPUTS:
        features (torch.Tensor) - the pixel embeddings, shape=(N, d) where
        N is the number of pixels and d is the embedding space dimension.

        labels (torch.Tensor) - ground-truth group labels, shape=(N, )

    OUTPUT:
        cluster_means (torch.Tensor) - (n_c, d) tensor where n_c is the number of
        distinct instances. Each row is a (1,d) vector corresponding to
        the coordinates of the i-th centroid.
    """
    group_ids = sorted(np.unique(labels).astype(int))
    cluster_means = []
    # print(group_ids)
    for c in group_ids:
        index = labels.astype(int) == c
        mu_c = features[index].mean(0)
        cluster_means.append(mu_c)
    cluster_means = np.vstack(cluster_means)
    return group_ids, cluster_means


def find_cluster_means_cuda(features, labels):
    r"""
    For a given image, compute the centroids \mu_c for each
    cluster label in the embedding space.

    INPUTS:
        features (torch.Tensor) - the pixel embeddings, shape=(N, d) where
        N is the number of pixels and d is the embedding space dimension.

        labels (torch.Tensor) - ground-truth group labels, shape=(N, )

    OUTPUT:
        cluster_means (torch.Tensor) - (n_c, d) tensor where n_c is the number of
        distinct instances. Each row is a (1,d) vector corresponding to
        the coordinates of the i-th centroid.
    """
    group_ids = torch.unique(labels).int()
    cluster_means = []
    # print(group_ids)
    for c in group_ids:
        index = labels.int() == c
        mu_c = torch.mean(features[index], dim=0)
        cluster_means.append(mu_c)
    cluster_means = torch.cat(cluster_means, dim=0)
    return group_ids, cluster_means


def fit_predict(
    embeddings, seediness, margins, fitfunc, s_threshold=0.0, p_threshold=0.5
):
    device = embeddings.device
    seediness = seediness.reshape(-1)
    num_points = embeddings.shape[0]
    count = 0
    probs = []
    unclustered = torch.ones(num_points, device=device, dtype=torch.bool)
    while count < num_points:
        seed_idx = (seediness * unclustered.float()).argmax()
        if seediness[seed_idx] < s_threshold:
            break
        centroid, sigma = embeddings[seed_idx], margins[seed_idx]
        f = fitfunc(centroid, sigma)
        p = f(embeddings)
        probs.append(p.view(-1, 1))
        count += torch.sum(p[unclustered] > p_threshold)
        unclustered[p > p_threshold] = False
    if len(probs) == 0:
        return torch.zeros(num_points, device=device).long()
    probs = torch.cat(probs, dim=1)
    labels = probs.argmax(dim=1)
    return labels


def fit_predict_np(
    embeddings, seediness, margins, fitfunc, s_threshold=0.0, p_threshold=0.5
):
    num_points = embeddings.shape[0]
    seediness = seediness.reshape(-1)
    count = 0
    probs = []
    unclustered = np.ones(num_points, dtype=np.bool_)
    while count < num_points:
        seed_idx = (seediness * unclustered.astype(float)).argmax()
        if seediness[seed_idx] < s_threshold:
            break
        centroid, sigma = embeddings[seed_idx], margins[seed_idx]
        f = fitfunc(centroid, sigma)
        p = f(embeddings)
        probs.append(p.reshape(-1, 1))
        count += np.sum(p[unclustered] > p_threshold)
        unclustered[p > p_threshold] = False
    if len(probs) == 0:
        return np.zeros(num_points).astype(np.long)
    probs = np.concatenate(probs, axis=1)
    labels = probs.argmax(axis=1)
    return labels


def fit_predict_verbose(
    embeddings, seediness, margins, fitfunc, s_threshold=0.0, p_threshold=0.5
):
    pred_labels = -np.ones(embeddings.shape[0])
    probs = []
    spheres = []
    seediness_copy = np.copy(seediness)
    count = 0
    # print(seediness.shape[0])
    while count < seediness.shape[0]:
        # print(count)
        i = np.argsort(seediness_copy)[::-1][0]
        seedScore = seediness[i]
        if seedScore < s_threshold:
            break
        centroid = embeddings[i]
        sigma = margins[i]
        spheres.append((centroid, sigma))
        f = fitfunc(centroid, sigma)
        pValues = f(embeddings)
        probs.append(pValues.reshape(-1, 1))
        cluster_index = np.logical_and((pValues > p_threshold), (seediness_copy > 0))
        # print(cluster_index)
        seediness_copy[cluster_index] = -1
        if sum(cluster_index) == 0:
            break
        count += sum(cluster_index)
    if len(probs) == 0:
        return pred_labels, spheres, 1, float("inf")
    cluster_count = len(probs)
    probs = np.hstack(probs)
    pred_labels = np.argmax(probs, axis=1)
    ll = np.sum(np.log(np.max(probs, axis=1) + 1e-8))
    # if cluster_all:
    #     pred_labels = cluster_remainder(embeddings, pred_labels)
    return pred_labels, spheres, cluster_count, ll


# ----------------VARIOUS POST-PROCESSING ROUTINES FOR ANALYSIS---------------


def main_loop_parameter_search(train_cfg, **kwargs):
    from spine.main_funcs import cycle

    inference_cfg = make_inference_cfg(
        train_cfg, gpu=kwargs["gpu"], batch_size=1, model_path=kwargs["model_path"]
    )
    start_index = kwargs.get("start_index", 0)
    end_index = kwargs.get("end_index", 20000)
    event_list = list(range(start_index, end_index))
    loader = loader_factory(inference_cfg, event_list=event_list)
    dataset = iter(cycle(loader))
    driver = Driver(inference_cfg)
    output = []

    inference_cfg["trainval"]["iterations"] = len(event_list)
    iterations = inference_cfg["trainval"]["iterations"]

    p_lims = tuple(kwargs["p_lims"])
    s_lims = tuple(kwargs["s_lims"])
    p_mesh = int(kwargs["p_mesh"])
    s_mesh = int(kwargs["s_mesh"])

    p_range = np.linspace(p_lims[0], p_lims[1], p_mesh)
    s_range = np.linspace(s_lims[0], s_lims[1], s_mesh)

    for i in event_list:

        print("Iteration: %d" % i)

        start = time.time()
        data_blob, res = driver.forward(dataset)
        end = time.time()
        forward_time = float(end - start)
        # segmentation = res['segmentation'][0]
        embedding = res["embeddings"][0]
        seediness = res["seediness"][0].reshape(
            -1,
        )
        margins = res["margins"][0].reshape(
            -1,
        )
        # print(data_blob['segment_label'][0])
        # print(data_blob['cluster_label'][0])
        if isinstance(data_blob["cluster_label"][0], torch.Tensor):
            semantic_labels = (
                data_blob["cluster_label"][0][:, -1].detach().cpu().numpy()
            )
            print(semantic_labels.shape)
            print(embedding.shape)
            cluster_labels = data_blob["cluster_label"][0][:, 5].detach().cpu().numpy()
            # print(data_blob['segment_label'][0])
            # print(semantic_labels)
            # print(np.unique(cluster_labels))
            coords = data_blob["input_data"][0][:, :3].detach().cpu().numpy()
        else:
            semantic_labels = data_blob["cluster_label"][0][:, -1]
            print(semantic_labels.shape)
            print(embedding.shape)
            cluster_labels = data_blob["cluster_label"][0][:, 5]
            # print(data_blob['segment_label'][0])
            # print(semantic_labels)
            # print(np.unique(cluster_labels))
            coords = data_blob["input_data"][0][:, :3]
        index = data_blob["index"][0]

        acc_dict = {}

        for p in p_range:
            for s in s_range:
                print("---------------------------------------------")
                print("p0 = {}, s0 = {}".format(p, s))
                for c in np.unique(semantic_labels):
                    if int(c) == 4:
                        continue
                    semantic_mask = semantic_labels == c
                    clabels = cluster_labels[semantic_mask]
                    voxel_counts = clabels.shape[0]
                    embedding_class = embedding[semantic_mask]
                    coords_class = coords[semantic_mask]
                    seed_class = seediness[semantic_mask]
                    margins_class = margins[semantic_mask]

                    start = time.time()
                    pred = fit_predict_np(
                        embedding_class,
                        seed_class,
                        margins_class,
                        gaussian_kernel,
                        s_threshold=s,
                        p_threshold=p,
                    )
                    end = time.time()
                    post_time = float(end - start)
                    purity, efficiency = purity_efficiency(pred, clabels)
                    fscore = 2 * (purity * efficiency) / (purity + efficiency)
                    ari = ARI(pred, clabels)
                    sbd = SBD(pred, clabels)
                    true_num_clusters = len(np.unique(clabels))
                    _, true_centroids = find_cluster_means(coords_class, clabels)
                    row = (
                        index,
                        c,
                        ari,
                        purity,
                        efficiency,
                        fscore,
                        sbd,
                        true_num_clusters,
                        s,
                        p,
                        forward_time,
                        post_time,
                        voxel_counts,
                    )
                    output.append(row)
                    print("ARI = ", ari)

    output = pd.DataFrame(
        output,
        columns=[
            "Index",
            "Class",
            "ARI",
            "Purity",
            "Efficiency",
            "FScore",
            "SBD",
            "true_num_clusters",
            "seed_threshold",
            "prob_threshold",
            "forward_time",
            "post_time",
            "voxel_counts",
        ],
    )
    return output


def main_loop(train_cfg, **kwargs):
    from spine.main_funcs import cycle

    inference_cfg = make_inference_cfg(
        train_cfg, gpu=kwargs["gpu"], batch_size=1, model_path=kwargs["model_path"]
    )
    start_index = kwargs.get("start_index", 0)
    end_index = kwargs.get("end_index", 20000)
    event_list = list(range(start_index, end_index))
    loader = loader_factory(inference_cfg, event_list=event_list)
    dataset = iter(cycle(loader))
    driver = Driver(inference_cfg)
    output = []

    inference_cfg["trainval"]["iterations"] = len(event_list)
    iterations = inference_cfg["trainval"]["iterations"]
    s_thresholds = kwargs["s_thresholds"]
    p_thresholds = kwargs["p_thresholds"]

    for i in event_list:

        print("Iteration: %d" % i)

        start = time.time()
        data_blob, res = driver.forward(dataset)
        end = time.time()
        forward_time = float(end - start)
        # segmentation = res['segmentation'][0]
        embedding = res["embeddings"][0]
        seediness = res["seediness"][0].reshape(
            -1,
        )
        margins = res["margins"][0].reshape(
            -1,
        )
        # print(data_blob['segment_label'][0])
        # print(data_blob['cluster_label'][0])
        if isinstance(data_blob["cluster_label"][0], torch.Tensor):
            semantic_labels = (
                data_blob["cluster_label"][0][:, -1].detach().cpu().numpy()
            )
            cluster_labels = data_blob["cluster_label"][0][:, 5].detach().cpu().numpy()
            # print(data_blob['segment_label'][0])
            # print(semantic_labels)
            # print(np.unique(cluster_labels))
            coords = data_blob["input_data"][0][:, :3].detach().cpu().numpy()
        else:
            semantic_labels = data_blob["cluster_label"][0][:, -1]
            cluster_labels = data_blob["cluster_label"][0][:, 5]
            # print(data_blob['segment_label'][0])
            # print(semantic_labels)
            # print(np.unique(cluster_labels))
            coords = data_blob["input_data"][0][:, :3]
        index = data_blob["index"][0]

        acc_dict = {}

        for c in np.unique(semantic_labels):
            if int(c) == 4:
                continue
            semantic_mask = semantic_labels == c
            clabels = cluster_labels[semantic_mask]
            embedding_class = embedding[semantic_mask]
            coords_class = coords[semantic_mask]
            seed_class = seediness[semantic_mask]
            margins_class = margins[semantic_mask]

            start = time.time()
            pred = fit_predict_np(
                embedding_class,
                seed_class,
                margins_class,
                gaussian_kernel,
                s_threshold=s_thresholds[int(c)],
                p_threshold=p_thresholds[int(c)],
            )
            end = time.time()
            post_time = float(end - start)
            # pred, spheres, cluster_count = fit_predict2(embedding_class, seed_class, margins_class, gaussian_kernel,
            #                     s_threshold=s_threshold, p_threshold=p_threshold, cluster_all=True)
            purity, efficiency = purity_efficiency(pred, clabels)
            fscore = 2 * (purity * efficiency) / (purity + efficiency)
            ari = ARI(pred, clabels)
            sbd = SBD(pred, clabels)
            true_num_clusters = len(np.unique(clabels))
            pred_num_clusters = len(np.unique(pred))
            row = (
                index,
                c,
                ari,
                purity,
                efficiency,
                fscore,
                sbd,
                true_num_clusters,
                pred_num_clusters,
                s_thresholds[int(c)],
                p_thresholds[int(c)],
                forward_time,
                post_time,
            )
            output.append(row)
            print("ARI = ", ari)
            print("SBD = ", sbd)
            # print("LL = ", ll)

    output = pd.DataFrame(
        output,
        columns=[
            "Index",
            "Class",
            "ARI",
            "Purity",
            "Efficiency",
            "FScore",
            "SBD",
            "true_num_clusters",
            "pred_num_clusters",
            "seed_threshold",
            "prob_threshold",
            "forward_time",
            "post_time",
        ],
    )
    return output


def main_loop_voxel_cut(train_cfg, **kwargs):
    from spine.main_funcs import cycle

    inference_cfg = make_inference_cfg(
        train_cfg, gpu=kwargs["gpu"], batch_size=1, model_path=kwargs["model_path"]
    )
    start_index = kwargs.get("start_index", 0)
    end_index = kwargs.get("end_index", 20000)
    event_list = list(range(start_index, end_index))
    loader = loader_factory(inference_cfg, event_list=event_list)
    dataset = iter(cycle(loader))
    driver = Driver(inference_cfg)
    output = []

    inference_cfg["trainval"]["iterations"] = len(event_list)
    iterations = inference_cfg["trainval"]["iterations"]
    s_thresholds = kwargs["s_thresholds"]
    p_thresholds = kwargs["p_thresholds"]
    # s_thresholds = {0: 0.88, 1: 0.92, 2: 0.84, 3: 0.84, 4: 0.8}
    # p_thresholds = {0: 0.5, 1: 0.5, 2: 0.5, 3: 0.5, 4: 0.5}
    # s_thresholds = {0: 0.65, 1: 0.95, 2: 0.25, 3: 0.85, 4: 0.0} # F32D6 Parameters
    # p_thresholds = {0: 0.31, 1: 0.21, 2: 0.06, 3: 0.26, 4: 0.11}
    # s_thresholds = { key : s_threshold for key in range(5)}
    # p_thresholds = { key : p_threshold for key in range(5)}

    for i in event_list:

        print("Iteration: %d" % i)

        start = time.time()
        data_blob, res = driver.forward(dataset)
        end = time.time()
        forward_time = float(end - start)
        # segmentation = res['segmentation'][0]
        embedding = res["embeddings"][0]
        seediness = res["seediness"][0].reshape(
            -1,
        )
        margins = res["margins"][0].reshape(
            -1,
        )
        # print(data_blob['segment_label'][0])
        # print(data_blob['cluster_label'][0])
        semantic_labels = data_blob["cluster_label"][0][:, -1]
        cluster_labels = data_blob["cluster_label"][0][:, 5]
        coords = data_blob["input_data"][0][:, :3]
        index = data_blob["index"][0]

        acc_dict = {}

        for c in np.unique(semantic_labels):
            if int(c) == 4:
                continue
            semantic_mask = semantic_labels == c
            clabels = cluster_labels[semantic_mask]
            embedding_class = embedding[semantic_mask]
            coords_class = coords[semantic_mask]
            seed_class = seediness[semantic_mask]
            margins_class = margins[semantic_mask]
            print(index, c)
            voxel_mask = np.ones(clabels.shape[0]).astype(bool)
            for j, cluster_id in enumerate(np.unique(clabels)):
                if sum(clabels == cluster_id) < 10:
                    voxel_mask[clabels == cluster_id] = False
            if sum(voxel_mask) < 10:
                continue
            start = time.time()
            pred, spheres, cluster_count, ll = fit_predict2(
                embedding_class[voxel_mask],
                seed_class[voxel_mask],
                margins_class[voxel_mask],
                gaussian_kernel,
                s_threshold=s_thresholds[int(c)],
                p_threshold=p_thresholds[int(c)],
            )
            end = time.time()
            post_time = float(end - start)
            # pred, spheres, cluster_count = fit_predict2(embedding_class, seed_class, margins_class, gaussian_kernel,
            #                     s_threshold=s_threshold, p_threshold=p_threshold, cluster_all=True)
            purity, efficiency = purity_efficiency(pred, clabels[voxel_mask])
            fscore = 2 * (purity * efficiency) / (purity + efficiency)
            ari = ARI(pred, clabels[voxel_mask])
            sbd = SBD(pred, clabels[voxel_mask])
            true_num_clusters = len(np.unique(clabels[voxel_mask]))
            _, true_centroids = find_cluster_means(
                coords_class[voxel_mask], clabels[voxel_mask]
            )
            for j, cluster_id in enumerate(np.unique(clabels[voxel_mask])):
                margin = np.mean(margins_class[clabels == cluster_id])
                true_size = np.std(
                    np.linalg.norm(
                        coords_class[clabels == cluster_id] - true_centroids[j], axis=1
                    )
                )
                voxel_count = sum(clabels == cluster_id)
                row = (
                    index,
                    c,
                    ari,
                    purity,
                    efficiency,
                    fscore,
                    sbd,
                    true_num_clusters,
                    cluster_count,
                    s_thresholds[int(c)],
                    p_thresholds[int(c)],
                    margin,
                    true_size,
                    forward_time,
                    post_time,
                    voxel_count,
                )
                output.append(row)
            print("ARI = ", ari)
            print("SBD = ", sbd)
            # print("LL = ", ll)

    output = pd.DataFrame(
        output,
        columns=[
            "Index",
            "Class",
            "ARI",
            "Purity",
            "Efficiency",
            "FScore",
            "SBD",
            "true_num_clusters",
            "pred_num_clusters",
            "seed_threshold",
            "prob_threshold",
            "margin",
            "true_size",
            "forward_time",
            "post_time",
            "voxel_count",
        ],
    )
    return output
